#!/usr/bin/env python
# control.py

"""
.. module:: dasystem
.. moduleauthor:: Wouter Peters 

Revision History:
File created on 26 Aug 2010.

The DaSystem class is found in the module :mod:`dasystem`, or in a specific implementation under the da/ source tree. It is derived from the standard python :class:`dictionary` object. 

It describes the details of the data assimilation system used (i.e., CarbonTracker, or CT Methane, or ....) ::

    datadir         : /Volumes/Storage/CO2/carbontracker/input/ct08/   ! The directory where input data is found
    obs.input.dir   : ${datadir}/obsnc/with_fillvalue                  ! the observation input dir
    obs.input.fname : obs_forecast.nc                                  ! the observation input file
    ocn.covariance  : ${datadir}/oif_p3_era40.dpco2.2000.01.hdf        ! the ocean flux covariance file
    bio.covariance  : ${datadir}/covariance_bio_olson19.nc             ! the biosphere flux covariance file
    deltaco2.prefix : oif_p3_era40.dpco2                               ! the type of ocean product used
    regtype         : olson19_oif30                                    ! the ecoregion definitions
    nparameters     : 240                                              ! the number of parameters to solve for
    random.seed     : 4385                                             ! the random seed for the first cycle
    regionsfile     : transcom_olson19_oif30.hdf                       ! the ecoregion defintion mask file

    ! Info on the sites file used

    obs.sites.rc        : ${datadir}/sites_and_weights_co2.ct10.rc     ! the weights in the covariance matric of each obs

The full baseclass description:

.. autoclass:: da.baseclasses.dasystem.DaSystem
   :members:

"""

import os
import sys
import logging
import datetime

################### Begin Class DaSystem ###################

class DaSystem(dict):
    """ 
    Information on the data assimilation system used. This is normally an rc-file with settings.
    """

    def __init__(self,rcfilename):
        """
        Initialization occurs from passed rc-file name, items in the rc-file will be added
        to the dictionary
        """

        self.Identifier     = 'CarbonTracker CO2'    # the identifier gives the platform name

        self.LoadRc(rcfilename)

        msg  = 'Data Assimilation System initialized: %s'%self.Identifier ; logging.debug(msg)


    def __str__(self):
        """
        String representation of a DaInfo object
        """

        msg = "==============================================================="    ; print msg
        msg = "DA System Info rc-file is %s" % self.RcFileName                                ; print msg
        msg = "==============================================================="    ; print msg

        return ""


    def LoadRc(self,RcFileName):
        """ 
        This method loads a DA System Info rc-file with settings for this simulation 
        """
        import da.tools.rc as rc 

        for k,v in rc.read(RcFileName).iteritems():
            self[k]             = v
        self.RcFileName     = RcFileName
        self.DaRcLoaded    = True

        msg                 = 'DA System Info rc-file (%s) loaded successfully'%self.RcFileName ; logging.debug(msg)

        return True

    def Initialize(self ):
        """ 
        Initialize the object. 
        """


    def Validate(self ):
        """ 
        Validate the contents of the rc-file given a dictionary of required keys
        """

        needed_rc_items={}

        for k,v in self.iteritems():
            if v == 'True' : self[k] = True
            if v == 'False': self[k] = False

        for key in needed_rc_items:

            if not self.has_key(key):
                status,msg = ( False,'Missing a required value in rc-file : %s' % key)
                logging.error(msg)
                raise IOError,msg

        status,msg = ( True,'DA System Info settings have been validated succesfully' )  ; logging.debug(msg)

        return None
################### End Class DaSystem ###################


if __name__ == "__main__":
    pass
