#!/usr/bin/env python
# obs.py

"""
.. module:: obs
.. moduleauthor:: Wouter Peters 

Revision History:
File created on 28 Jul 2010.

.. autoclass:: da.baseclasses.obs.Observation 
   :members: Initialize, Validate, AddObs, AddSimulations, AddModelDataMismatch, WriteSampleInfo  

.. autoclass:: da.baseclasses.obs.ObservationList 
   :members: __init__

"""
import os
import sys
import logging
import datetime
from numpy import array, ndarray

identifier = 'Observation baseclass'
version    = '0.0'

################### Begin Class Observation ###################

class Observation(object):
    """ 
    The baseclass Observation is a generic object that provides a number of methods required for any type of observations used in 
    a data assimilation system. These methods are called from the CarbonTracker pipeline. 

    .. note:: Most of the actual functionality will need to be provided through a derived Observations class with the methods 
              below overwritten. Writing your own derived class for Observations is one of the first tasks you'll likely 
              perform when extending or modifying the CarbonTracker Data Assimilation Shell.

    Upon initialization of the class, an object is created that holds no actual data, but has a placeholder attribute `self.Data` 
    which is an empty list of type :class:`~da.baseclasses.obs.ObservationList`. An ObservationList object is created when the 
    method :meth:`~da.baseclasses.obs.Observation.AddObs` is invoked in the pipeline. 

    From the list of observations, a file is written  by method 
    :meth:`~da.baseclasses.obs.Observation.WriteSampleInfo`
    with the sample info needed by the 
    :class:`~da.baseclasses.observationoperator.ObservationOperator` object. The values returned after sampling 
    are finally added by :meth:`~da.baseclasses.obs.Observation.AddSimulations`

    """ 

    def __init__(self, DaCycle = None):
        """
        create an object with an identifier, version, and an empty ObservationList
        """
        self.Identifier = self.getid()
        self.Version    = self.getversion()
        self.Data       = ObservationList([])  # Initialize with an empty list of obs

        # The following code allows the object to be initialized with a DaCycle object already present. Otherwise, it can
        # be added at a later moment.

        if DaCycle != None:
            self.DaCycle = DaCycle
        else:
            self.DaCycle = {}

        msg                 = 'Observation object initialized: %s'%self.Identifier ; logging.info(msg)

    def getid(self):
        """
        Return the identifier of the Observation Object, used for logging purposed
        """
        return identifier

    def getversion(self):
        """
        Return the version of the Observation Object, used for logging purposed
        """
        return identifier

    def __str__(self):
        """ Prints a list of Observation objects"""
        return "This is a %s object, version %s"%(self.Identifier,self.Version)

    def Initialize(self):
        """ Perform all steps needed to start working with observational data, this can include moving data, concatenating files,
            selecting datasets, etc.
        """

    def Validate(self):
        """ Make sure that data needed for the ObservationOperator (such as observation input lists, or parameter files)
            are present.
        """

    def AddObs(self):
        """ 
        Add actual observation data to the Observation object. This is in a form of an 
        :class:`~da.baseclasses.obs.ObservationList` that is contained in self.Data. The 
        list has as only requirement that it can return the observed+simulated values 
        through the method :meth:`~da.baseclasses.obs.ObservationList.getvalues`

        """

    def AddSimulations(self):
        """ Add the simulation data to the Observation object. 
        """

    def AddModelDataMismatch(self ):
        """ 
            Get the model-data mismatch values for this cycle.

        """

    def WriteSampleInfo(self ):
        """ 
            Write the information needed by the observation operator to a file. Return the filename that was written for later use
        """



################### End Class Observations ###################

################### Begin Class ObservationList ###################

class ObservationList(list):
    """ This is a special type of list that holds observed sample objects. It has methods to extract data from such a list easily """

    def getvalues(self,name,constructor=array):

        result = constructor([getattr(o,name) for o in self])
        if isinstance(result,ndarray): 
            return result.squeeze()
        else:
            return result

################### End Class ObservationList ###################

