#!/usr/bin/env python
# ct_statevector_tools.py

"""
.. module:: statevector
.. moduleauthor:: Wouter Peters 

Revision History:
File created on 28 Jul 2010.

The module statevector implements the data structure and methods needed to work with state vectors (a set of unknown parameters to be optimized by a DA system) of different lengths, types, and configurations. Two baseclasses together form a generic framework:
    * :class:`~da.baseclasses.statevector.StateVector`
    * :class:`~da.baseclasses.statevector.EnsembleMember`

As usual, specific implementations of StateVector objects are done through inheritance form these baseclasses. An example of designing 
your own baseclass StateVector we refer to :ref:`tut_chapter5`.

.. autoclass:: da.baseclasses.statevector.StateVector 

.. autoclass:: da.baseclasses.statevector.EnsembleMember 

"""

import os
import sys
import logging
import numpy as np
import da.tools.io4 as io
import datetime

identifier = 'Baseclass Statevector '
version    = '0.0'

################### Begin Class EnsembleMember ###################

class EnsembleMember(object):
    """ 
        An ensemble member object consists of:
           * a member number
           * parameter values
           * an observation object to hold sampled values for this member

        Ensemble members are initialized by passing only an ensemble member number, all data is added by methods 
        from the :class:`~da.baseclasses.statevector.StateVector`. Ensemble member objects have almost no functionality 
        except to write their data to file using method :meth:`~da.baseclasses.statevector.EnsembleMember.WriteToFile`

        .. automethod:: da.baseclasses.statevector.EnsembleMember.__init__ 
        .. automethod:: da.baseclasses.statevector.EnsembleMember.WriteToFile 
        .. automethod:: da.baseclasses.statevector.EnsembleMember.AddCustomFields 

    """

    def __init__(self, membernumber):
        """
           :param memberno: integer ensemble number
           :rtype: None

           An EnsembleMember object is initialized with only a number, and holds two attributes as containter for later
           data:
                * ParameterValues, will hold the actual values of the parameters for this data
                * ModelSample, will hold an :class:`~da.baseclasses.obs.Observation` object and the model samples resulting from this members' data

        """

        self.membernumber           = membernumber   # the member number
        self.ParameterValues        = None           # Parameter values of this member

    def __str__(self):
        return "%03d"%self.membernumber

################### End Class EnsembleMember ###################

################### Begin Class StateVector ###################

import numpy as np
class StateVector(object):
    """ 
    The StateVector object first of all contains the data structure of a statevector, defined by 3 attributes that define the 
    dimensions of the problem in parameter space:
        * nlag
        * nparameters
        * nmembers

    The fourth important dimension `nobs` is not related to the StateVector directly but is initialized to 0, and later on 
    modified to be used in other parts of the pipeline:
        * nobs

    These values are set as soon as the :meth:`~da.baseclasses.statevector.StateVector.Initialize` is called from the :ref:`pipeline`. 
    Additionally, the value of attribute `isOptimized` is set to `False` indicating that the StateVector holds a-priori values 
    and has not been modified by the :ref:`optimizer`.

    StateVector objects can be filled with data in two ways
        1. By reading the data from file
        2. By creating the data through a set of method calls

    Option (1) is invoked using method :meth:`~da.baseclasses.statevector.StateVector.ReadFromFile`. 
    Option (2) consists of a call to method :meth:`~da.baseclasses.statevector.StateVector.MakeNewEnsemble`

    Each option makes use of the same call to :meth:`~da.baseclasses.statevector.StateVector.GetNewMember`, to create a 
    data container to be filled: the :class:`~da.baseclasses.statevector.EnsembleMember`.

    Once the StateVector object has been filled with data, it is used in the pipeline and a few more methods are
    invoked from there:
        * :meth:`~da.baseclasses.statevector.StateVector.Propagate`, to advance the StateVector from t=t to t=t+1
        * :meth:`~da.baseclasses.statevector.StateVector.WriteToFile`, to write the StateVector to a NetCDF file for later use

    The methods are described below:

    .. automethod:: da.baseclasses.statevector.StateVector.Initialize 
    .. automethod:: da.baseclasses.statevector.StateVector.ReadFromFile
    .. automethod:: da.baseclasses.statevector.StateVector.WriteToFile
    .. automethod:: da.baseclasses.statevector.StateVector.MakeNewEnsemble
    .. automethod:: da.baseclasses.statevector.StateVector.GetNewMember
    .. automethod:: da.baseclasses.statevector.StateVector.Propagate
    .. automethod:: da.baseclasses.statevector.StateVector.WriteMembersToFile

    Finally, the StateVector can be mapped to a gridded array, or to a vector of TransCom regions, using:

    .. automethod:: da.baseclasses.statevector.StateVector.GridToVector
    .. automethod:: da.baseclasses.statevector.StateVector.VectorToGrid
    .. automethod:: da.baseclasses.statevector.StateVector.VectorToTC
    .. automethod:: da.baseclasses.statevector.StateVector.StateToTC

    """

    def __init__(self, DaCycle = None):
        self.Identifier = self.getid()
        self.Version    = self.getversion()

        # The following code allows the object to be initialized with a DaCycle object already present. Otherwise, it can
        # be added at a later moment.

        if DaCycle != None:
            self.DaCycle = DaCycle
        else:
            self.DaCycle = {}


        msg                 = 'Statevector object initialized: %s'%self.Identifier ; logging.info(msg)

    def getid(self):
        return identifier

    def getversion(self):
        return version

    def Initialize(self):
        """
        Initialize the object by specifying the dimensions. 
        There are two major requirements for each statvector that you want to build:
        
            (1) is that the statevector can map itself onto a regular grid
            (2) is that the statevector can map itself (mean+covariance) onto TransCom regions

        An example is given below.
        """
        import da.tools.io4 as io
        import numpy as np


        dims                = ( int(self.DaCycle['time.nlag']),
                            int(self.DaCycle['da.optimizer.nmembers']),
                            int(self.DaCycle.DaSystem['nparameters']),
                           )


        self.nlag               = dims[0]
        self.nmembers           = dims[1]
        self.nparams            = dims[2]
        self.nobs               = 0
        self.isOptimized        = False
        self.ObsToAssimmilate   = ()  # empty containter to hold observations to assimilate later on

        # These list objects hold the data for each time step of lag in the system. Note that the ensembles for each time step consist 
        # of lists of EnsembleMember objects, we define member 0 as the mean of the distribution and n=1,...,nmembers as the spread.

        self.EnsembleMembers    = range(self.nlag)

        for n in range(self.nlag):
            self.EnsembleMembers[n] = []


        # This specifies the file to read with the gridded mask at 1x1 degrees. Each gridbox holds a number that specifies the parametermember
        #  that maps onto it. From this map, a dictionary is created that allows a reverse look-up so that we can map parameters to a grid.

        mapfile         = os.path.join(self.DaCycle.DaSystem['regionsfile'])
        ncf             = io.CT_Read(mapfile,'read')
        self.gridmap    = ncf.GetVariable('regions')
        self.tcmap      = ncf.GetVariable('transcom_regions')
        dummy           = ncf.close()

        msg = "A TransCom  map on 1x1 degree was read from file %s"%self.DaCycle.DaSystem['regionsfile'] ; logging.debug(msg)
        msg = "A parameter map on 1x1 degree was read from file %s"%self.DaCycle.DaSystem['regionsfile'] ; logging.debug(msg)

        # Create a dictionary for state <-> gridded map conversions

        nparams         = self.gridmap.max()
        self.griddict   = {}
        for r in range(1,nparams+1):
             sel    = (self.gridmap.flat == r).nonzero()
             if len(sel[0])>0: 
                 self.griddict[r]=sel

        msg = "A dictionary to map grids to states and vice versa was created" ; logging.debug(msg)

        # Create a matrix for state <-> TransCom conversions

        self.tcmatrix  = np.zeros((self.nparams,23,),'float') 

        for r in range(1,self.nparams+1):
             sel    = (self.gridmap.flat == r).nonzero()
             if len(sel[0])<1: 
                 continue
             else:
                 n_tc = set(self.tcmap.flatten().take(sel[0]))
                 if len(n_tc) > 1: 
                     msg  = "Parameter %d seems to map to multiple TransCom regions (%s), I do not know how to handle this" % (r,n_tc) ;logging.error(msg) 
                     raise ValueError
                 self.tcmatrix[r-1,n_tc.pop()-1]= 1.0

        msg = "A matrix to map states to TransCom regions and vice versa was created" ; logging.debug(msg)

        # Create a mask for species/unknowns

        dummy = self.MakeSpeciesMask()

    def __str__(self):
        return "This is a base class derived state vector object"

    def MakeSpeciesMask(self):

        """

        This method creates a dictionary with as key the name of a tracer, and as values an array of 0.0/1.0 values 
        specifying which StateVector elements are constrained by this tracer. This mask can be used in 
        the optimization to ensure that certain types of osbervations only update certain unknowns.

        An example would be that the tracer '14CO2' can be allowed to only map onto fossil fuel emissions in the state

        The form of the mask is:

        {'co2': np.ones(self.nparams), 'co2c14', np.zeros(self.nparams)  }

        so that 'co2' maps onto all parameters, and 'co2c14' on none at all. These arrays are used in the Class 
        optimizer when state updates are actually performed

        """

        self.speciesdict = {'co2': np.ones(self.nparams)}

        msg = "A species mask was created, only the following species are recognized in this system:" ; logging.debug(msg)
        for k in self.speciesdict.keys(): 
            msg = "   ->    %s"%k ; logging.debug(msg)

        return None


    def MakeNewEnsemble(self,lag, covariancematrix = None):
        """ 
        :param lag: an integer indicating the time step in the lag order
        :param covariancematrix: a matrix to draw random values from
        :rtype: None
    
        Make a new ensemble, the attribute lag refers to the position in the state vector. 
        Note that lag=1 means an index of 0 in python, hence the notation lag-1 in the indexing below.
        The argument is thus referring to the lagged state vector as [1,2,3,4,5,..., nlag]

        The optional covariance object to be passed holds a matrix of dimensions [nparams, nparams] which is
        used to draw ensemblemembers from. If this argument is not passed it will ne substituted with an 
        identity matrix of the same dimensions.

        """    

        if covariancematrix == None: 
            covariancematrix=np.identity(self.nparams)

        # Make a cholesky decomposition of the covariance matrix


        U,s,Vh  = np.linalg.svd(covariancematrix)
        dof     = np.sum(s)**2/sum(s**2)
        C       = np.linalg.cholesky(covariancematrix)

        msg =   'Cholesky decomposition has succeeded '                    ; logging.debug(msg)
        msg =   'Appr. degrees of freedom in covariance matrix is %s'%(int(dof))  ; logging.info(msg)


        # Create mean values 

        NewMean                     = np.ones(self.nparams,float) # standard value for a new time step is 1.0

        # If this is not the start of the filter, average previous two optimized steps into the mix

        if lag == self.nlag and self.nlag >= 3:
            NewMean                     += self.EnsembleMembers[lag-2][0].ParameterValues + \
                                           self.EnsembleMembers[lag-3][0].ParameterValues 
            NewMean                      = NewMean/3.0

        # Create the first ensemble member with a deviation of 0.0 and add to list

        NewMember                   = self.GetNewMember(0)
        NewMember.ParameterValues   = NewMean.flatten()  # no deviations
        dummy                       = self.EnsembleMembers[lag-1].append(NewMember)

        # Create members 1:nmembers and add to EnsembleMembers list

        for member in range(1,self.nmembers):
            randstate                   = np.random.get_state()
            rands                       = np.random.randn(self.nparams)

            NewMember                   = self.GetNewMember(member)
            NewMember.ParameterValues   = np.dot(C,rands) + NewMean
            dummy                       = self.EnsembleMembers[lag-1].append(NewMember)

        msg =   '%d new ensemble members were added to the state vector # %d'%(self.nmembers,lag)  ; logging.debug(msg)

    def GetNewMember(self,memberno):
        """ 
        :param memberno: an integer indicating the ensemble member number
        :rtype: an empty ensemblemember object
        
        Return an ensemblemember object
        """ 

        return EnsembleMember(memberno)


    def Propagate(self):
        """ 
        :rtype: None

        Propagate the parameter values in the StateVector to the next cycle. This means a shift by one cycle 
        step for all states that will
        be optimized once more, and the creation of a new ensemble for the time step that just 
        comes in for the first time (step=nlag). 
        In the future, this routine can incorporate a formal propagation of the statevector.

        """
        from datetime import timedelta
        # Remove State Vector n=1 by simply "popping" it from the list and appending a new empty list at the front. This empty list will
        # hold the new ensemble for the new cycle 

        dummy = self.EnsembleMembers.pop(0)

        dummy = self.EnsembleMembers.append([])

        # And now create a new time step of mean + members for n=nlag

    
        date = self.DaCycle['time.start']+timedelta(days = (self.nlag-0.5)* int(self.DaCycle['time.cycle']))
        cov   = self.GetCovariance(date)
        dummy = self.MakeNewEnsemble(self.nlag,cov)

        msg   =   'The state vector has been propagated by one cycle ' ; logging.info(msg)

        return None

    def WriteToFile(self,filename):
        """
        :param filename: the full filename for the output NetCDF file
        :rtype: None

        Write the StateVector information to a NetCDF file for later use. 
        In principle the output file will have only one two datasets inside 
        called:
            * `meanstate`, dimensions [nlag, nparamaters]
            * `ensemblestate`, dimensions [nlag,nmembers, nparameters]

        This NetCDF information can be read back into a StateVector object using 
        :meth:`~da.baseclasses.statevector.StateVector.ReadFromFile`

        """
        #import da.tools.io4 as io
        #import da.tools.io as io
        import numpy as np

        if not self.isOptimized:
            f           = io.CT_CDF(filename,method='create')
            msg         = 'Creating new StateVector output file (%s)' % filename ; logging.debug(msg)
            qual        = 'prior'
        else:
            f           = io.CT_CDF(filename,method='write')
            msg         = 'Opening existing StateVector output file (%s)' % filename ; logging.debug(msg)
            qual        = 'opt'

        dimparams       = f.AddParamsDim(self.nparams)
        dimmembers      = f.AddMembersDim(self.nmembers)
        dimlag          = f.AddLagDim(self.nlag,unlimited=True)

        for n in range(self.nlag):

            members   = self.EnsembleMembers[n]
            MeanState =  members[0].ParameterValues

            savedict                = f.StandardVar(varname='meanstate_%s'%qual)
            savedict['dims']        = dimlag+dimparams 
            savedict['values']      = MeanState
            savedict['count']       = n
            savedict['comment']     = 'this represents the mean of the ensemble'
            dummy                   = f.AddData(savedict)

            members =  self.EnsembleMembers[n]
            devs    = np.asarray([m.ParameterValues.flatten() for m in members])
            data    = devs- np.asarray(MeanState)

            savedict                = f.StandardVar(varname='ensemblestate_%s'%qual)
            savedict['dims']        = dimlag+dimmembers+dimparams 
            savedict['values']      = data
            savedict['count']       = n
            savedict['comment']     = 'this represents deviations from the mean of the ensemble'
            dummy                   = f.AddData(savedict)

        dummy   = f.close()

        msg     = 'Successfully wrote the State Vector to file (%s) ' % (filename,) ; logging.info(msg)

    def ReadFromFile(self,filename,qual='opt'):
        """ 
        :param filename: the full filename for the input NetCDF file
        :param qual: a string indicating whether to read the 'prior' or 'opt'(imized) StateVector from file
        :rtype: None

        Read the StateVector information from a NetCDF file and put in a StateVector object
        In principle the input file will have only one four datasets inside 
        called:
            * `meanstate_prior`, dimensions [nlag, nparamaters]
            * `ensemblestate_prior`, dimensions [nlag,nmembers, nparameters]
            * `meanstate_opt`, dimensions [nlag, nparamaters]
            * `ensemblestate_opt`, dimensions [nlag,nmembers, nparameters]

        This NetCDF information can be written to file using 
        :meth:`~da.baseclasses.statevector.StateVector.WriteToFile`

        """

        #import da.tools.io as io
        import numpy as np

        f               = io.CT_Read(filename,'read')
        MeanState       = f.GetVariable('statevectormean_'+qual)
        EnsembleMembers = f.GetVariable('statevectorensemble_'+qual)
        dummy           = f.close()

        for n in range(self.nlag):
            if not self.EnsembleMembers[n] == []:
                self.EnsembleMembers[n] = []
                msg = 'Existing ensemble for lag=%d was removed to make place for newly read data'%(n+1,) ; logging.warning(msg)

            for m in range(self.nmembers):
                NewMember                   = self.GetNewMember(m)
                NewMember.ParameterValues   = EnsembleMembers[n,m,:].flatten() + MeanState[n]  # add the mean to the deviations to hold the full parameter values
                dummy                       = self.EnsembleMembers[n].append(NewMember)

        msg     = 'Successfully read the State Vector from file (%s) ' % (filename,) ; logging.info(msg)

    def WriteMembersToFile(self, lag):
        """ 
           :param: lag: Which lag step of the filter to write, must lie in range [1,...,nlag]
           :rtype: None

           Write ensemble member information to a NetCDF file for later use. The standard output filename is 
           *parameters.DDD.nc* where *DDD* is the number of the ensemble member. Standard output file location 
           is the `dir.input` of the DaCycle object. In principle the output file will have only two datasets inside 
           called `parametervalues` which is of dimensions `nparameters` and `parametermap` which is of dimensions (180,360). 
           This dataset can be read and used by a :class:`~da.baseclasses.observationoperator.ObservationOperator` object. 

           .. note:: if more, or other information is needed to complete the sampling of the ObservationOperator you
                     can simply inherit from the StateVector baseclass and overwrite this WriteMembersToFile function.

        """

        # These import statements caused a crash in netCDF4 on MacOSX. No problems on Jet though. Solution was
        # to do the import already at the start of the module, not just in this method.
           
        #import da.tools.io as io
        #import da.tools.io4 as io

        outdir  = self.DaCycle['dir.input']

        members =  self.EnsembleMembers[lag-1]

        for mem in members:

            filename        = os.path.join(outdir,'parameters.%03d.nc'% mem.membernumber)
            ncf             = io.CT_CDF(filename,method='create')
            dimparams       = ncf.AddParamsDim(self.nparams)
            dimgrid         = ncf.AddLatLonDim()

            data            =  mem.ParameterValues

            savedict                = io.std_savedict.copy()
            savedict['name']        = "parametervalues"
            savedict['long_name']   = "parameter_values_for_member_%d"%mem.membernumber
            savedict['units']       = "unitless"
            savedict['dims']        = dimparams 
            savedict['values']      = data
            savedict['comment']     = 'These are parameter values to use for member %d'%mem.membernumber
            dummy                   = ncf.AddData(savedict)

            griddata        = self.VectorToGrid(vectordata=data)


            savedict                = io.std_savedict.copy()
            savedict['name']        = "parametermap"
            savedict['long_name']   = "parametermap_for_member_%d"%mem.membernumber
            savedict['units']       = "unitless"
            savedict['dims']        = dimgrid 
            savedict['values']      = griddata.tolist()
            savedict['comment']     = 'These are gridded parameter values to use for member %d'%mem.membernumber
            dummy                   = ncf.AddData(savedict)

            dummy                   = ncf.close()

            msg     = 'Successfully wrote data from ensemble member %d to file (%s) ' % (mem.membernumber,filename,) ; logging.debug(msg)

    def GridToVector(self,griddata=None, method = 'avg'):
        """ 
            Map gridded data onto a vector of length (nparams,)

           :param griddata: a gridded dataset to use. This dataset is mapped onto a vector of length `nparams`
           :param method: a string that specifies the method to combine grid boxes in case reverse=True. Must be either ['avg','sum','minval']
           :rtype: ndarray: size (nparameters,)

           This method makes use of a dictionary that links every parameter number [1,...,nparams] to a series of gridindices. These 
           indices specify a location on a 360x180 array, stretched into a vector using `array.flat`. There are multiple ways of calling 
           this method::

               values       = self.GridToVector(griddata=mygriddeddata,method='minval') # 
                                                                    using the minimum value of all datapoints covered by that parameter index

               values       = self.GridToVector(griddata=mygriddeddata,method='avg') # 
                                                                    using the average value of all datapoints covered by that parameter index

               values       = self.GridToVector(griddata=mygriddeddata,method='sum') # 
                                                                    using the sum of values of all datapoints covered by that parameter index

           .. note:: This method uses a DaSystem object that must be initialized with a proper parameter map. See :class:`~da.baseclasses.dasystem` for details

        """

        methods =  ['avg','sum','minval']
        if method not in methods:
            msg = "To put data from a map into the statevector, please specify the method to use (%s)" % methods ; logging.error(msg)
            raise ValueError

        result=np.zeros((self.nparams,),float)
        for k,v in self.griddict.iteritems():
            #print k,k-1,result.shape, v
            if method == "avg": 
                result[k-1]=griddata.take(v).mean()
            elif method == "sum" : 
                result[k-1]=griddata.take(v).sum()
            elif method == "minval" : 
                result[k-1]=griddata.take(v).min()
        return result # Note that the result is returned, but not yet placed in the member.ParameterValues attrtibute!


    def VectorToGrid(self,vectordata=None):
        """ 
            Map vector elements to a map or vice cersa

           :param vectordata: a vector dataset to use in case `reverse = False`. This dataset is mapped onto a 1x1 grid and must be of length `nparams`
           :rtype: ndarray: an array of size (360,180,) 

           This method makes use of a dictionary that links every parameter number [1,...,nparams] to a series of gridindices. These 
           indices specify a location on a 360x180 array, stretched into a vector using `array.flat`. There are multiple ways of calling 
           this method::

               griddedarray = self.VectorToGrid(vectordata=ParameterValues) # simply puts the ParameterValues onto a (180,360,) array

           .. note:: This method uses a DaSystem object that must be initialzied with a proper parameter map. See :class:`~da.baseclasses.dasystem` for details

        """

        result=np.zeros(self.gridmap.shape,float)
        for k,v in self.griddict.iteritems():
            #print k,v
            result.put(v,vectordata[k-1])

        return result         

    def VectorToTC(self,vectordata,cov=False):
        """ 
            project Vector onto TransCom regions 

           :param vectordata: a vector dataset to use, must be of length `nparams`
           :param cov: a Boolean to specify whether the input dataset is a vector (mean), or a matrix (covariance)
           :rtype: ndarray: an array of size (23,) (cov:F) or of size (23,23,) (cov:T)
        """

        M = self.tcmatrix
        if cov:
            return np.dot(np.transpose(M), np.dot(vectordata,M) )
        else:
            return np.dot(vectordata.squeeze(),M)

    def StateToGrid(self, fluxvector = None, lag=1):
        """ 
            Transforms the StateVector information (mean + covariance) to a 1x1 degree grid.

            :param: fluxvector: a vector of length (nparams,) that holds the fluxes associated with each parameter in the StateVector
            :param: lag: the lag at which to evaluate the StateVector
            :rtype: a tuple of two arrays (gridmean,gridvariance) with dimensions (180,360,)

            If the attribute `fluxvector` is not passed, the function will return the mean parameter value and its variance on a 1x1 map.
            
            ..note:: Although we can return the variance information for each gridbox, the covariance information contained in the original ensemble is lost when mapping to 1x1 degree!

        """

        if fluxvector == None:
            fluxvector = np.ones(self.nparams)

        ensemble = self.EnsembleMembers[lag-1]

        ensemblemean =ensemble[0].ParameterValues

        # First transform the mean

        gridmean = self.VectorToGrid(vectordata = ensemblemean*fluxvector)

        # And now the covariance, first create covariance matrix (!), and then multiply

        deviations  = np.array([mem.ParameterValues*fluxvector-ensemblemean for mem in ensemble])
        variance    = deviations.std(axis=0)**2
        gridvar     = self.VectorToGrid(variance)

        return (gridmean,gridvar,)

    def StateToTC(self, fluxvector = None, lag=1):
        """ 
            Transforms the StateVector information (mean + covariance) to the TransCom regions.

            :param: fluxvector: a vector of length (nparams,) that holds the fluxes associated with each parameter in the StateVector
            :param: lag: the lag at which to evaluate the StateVector
            :rtype: a tuple of two arrays (mean,covariance) with dimensions ((23,), (23,23,) )

        """

        ensemble = self.EnsembleMembers[lag-1]

        ensemblemean =ensemble[0].ParameterValues

        # First transform the mean

        mean = self.VectorToTC(vectordata = ensemble[0].ParameterValues*fluxvector)

        # And now the covariance, first create covariance matrix (!), and then multiply

        deviations = np.array([mem.ParameterValues*fluxvector-ensemblemean for mem in ensemble])
        covariance = np.dot(np.transpose(deviations),deviations)/(self.nmembers-1)
        cov        = self.VectorToTC(covariance,cov=True)

        return (mean,cov,)


################### End Class StateVector ###################

if __name__ == "__main__":

    sys.path.append('../../')

    import os
    import sys
    from da.tools.general import StartLogger 
    from da.tools.initexit import CycleControl 
    import numpy as np
    import da.tools.rc as rc

    opts = ['-v']
    args = {'rc':'../../da.rc','logfile':'da_initexit.log','jobrcfilename':'test.rc'}

    StartLogger()
    DaCycle = CycleControl(opts,args)

    dummy = EnsembleMember(0)
    print dummy

    DaCycle.Initialize()
    print DaCycle

    dims        = ( int(DaCycle['time.nlag']),
                  int(DaCycle['da.optimizer.nmembers']),
                  int(DaCycle.DaSystem['nparameters']),
                  )

    StateVector = StateVector(dims)

    StateVector.MakeNewEnsemble(lag=1)

    members =  StateVector.EnsembleMembers[0]

    members[0].WriteToFile(DaCycle['dir.input'])

    StateVector.Propagate()

    savedir         = DaCycle['dir.output']
    filename        = os.path.join(savedir,'savestate.nc')

    StateVector.WriteToFile(filename)


    savedir         = DaCycle['dir.output']
    filename        = os.path.join(savedir,'savestate.nc')

    StateVector.ReadFromFile(filename)

