#!/usr/bin/env python
# ct_statevector_tools.py

"""
Author : peters 

Revision History:
File created on 28 Jul 2010.

"""

import os
import sys
sys.path.append(os.getcwd())

import logging
import datetime
from da.baseclasses.statevector import EnsembleMember, StateVector
import numpy as np

identifier = 'CarbonTracker Statevector '
version    = '0.0'

################### Begin Class CtStateVector ###################

class CtStateVector(StateVector):
    """ This is a StateVector object for CarbonTracker. It has a private method to make new ensemble members """

    def getid(self):
        return identifier

    def getversion(self):
        return version

    def GetCovariance(self,date):
        """ Make a new ensemble from specified matrices, the attribute lag refers to the position in the state vector. 
            Note that lag=1 means an index of 0 in python, hence the notation lag-1 in the indexing below.
            The argument is thus referring to the lagged state vector as [1,2,3,4,5,..., nlag]
        """    

        import da.tools.io4 as io
        try:
            import matplotlib.pyplot as plt
        except:
            pass

        # Get the needed matrices from the specified covariance files

        file_ocn_cov = self.DaCycle.DaSystem['ocn.covariance'] 
        file_bio_cov = self.DaCycle.DaSystem['bio.covariance'] 

        # replace YYYY.MM in the ocean covariance file string

        file_ocn_cov = file_ocn_cov.replace('2000.01',date.strftime('%Y.%m'))

        for file in [file_ocn_cov,file_bio_cov]:

            if not os.path.exists(file):

                msg = "Cannot find the specified file %s" % file ; logging.error(msg)
                raise IOError,msg
            else:

                msg = "Using covariance file: %s" % file ; logging.info(msg)

        f_ocn   = io.CT_Read(file_ocn_cov,'read')
        f_bio   = io.CT_Read(file_bio_cov,'read')

        cov_ocn = f_ocn.GetVariable('CORMAT')
        if f_bio.variables.has_key('covariance'):
            cov_bio = f_bio.GetVariable('covariance')  # newly created CTDAS covariance files
        else:
            cov_bio = f_bio.GetVariable('qprior')  # old CarbonTracker covariance files

        dummy   = f_ocn.close()
        dummy   = f_bio.close()

        dummy   = logging.debug("Succesfully closed files after retrieving prior covariance matrices")

        # Once we have the matrices, we can start to make the full covariance matrix, and then decompose it

        fullcov = np.zeros((self.nparams,self.nparams),float)

        nocn    = cov_ocn.shape[0]
        nbio    = cov_bio.shape[0]

        fullcov[0:nbio,0:nbio]                  = cov_bio
        fullcov[nbio:nbio+nocn,nbio:nbio+nocn]  = cov_ocn
        fullcov[nocn+nbio,nocn+nbio]            = 1.e-10


        try:
            plt.imshow(fullcov)
            plt.colorbar()
            plt.savefig('fullcovariancematrix.png')
            plt.close('all')
            dummy   = logging.debug("Covariance matrix visualized for inspection")
        except:
            pass

        return fullcov


################### End Class CtStateVector ###################


if __name__ == "__main__":

    import os
    import sys
    from da.tools.initexit import StartLogger 
    from da.tools.pipeline import JobStart
    import numpy as np

    sys.path.append(os.getcwd())

    opts = ['-v']
    args = {'rc':'da.rc','logfile':'da_initexit.log','jobrcfilename':'test.rc'}

    StartLogger()

    DaCycle = JobStart(opts,args)

    DaCycle.Initialize()

    StateVector = CtStateVector()

    StateVector.Initialize()

    for n in range(dims[0]):
        cov   = StateVector.GetCovariance()
        dummy = StateVector.MakeNewEnsemble(n+1,cov)

    StateVector.Propagate()

    savedir         = DaCycle['dir.output']
    filename        = os.path.join(savedir,'savestate.nc')

    dummy = StateVector.WriteToFile()

    StateVector.ReadFromFile(filename)

