#!/usr/bin/env python
# ct_statevector_tools.py

"""
Author : peters 

Revision History:
File created on 28 Jul 2010.

"""

import os
import sys
sys.path.append(os.getcwd())
sys.path.append('../../')


import logging
import datetime
from da.baseclasses.statevector import EnsembleMember, StateVector
import numpy as np

identifier = 'CarbonTracker Gridded Statevector '
version    = '0.0'

################### Begin Class CtStateVector ###################

class CtGriddedStateVector(StateVector):
    """ This is a StateVector object for CarbonTracker. It has a private method to make new ensemble members """

    def getid(self):
        return identifier

    def getversion(self):
        return version

    def GetCovariance(self,date):
        """ Make a new ensemble from specified matrices, the attribute lag refers to the position in the state vector. 
            Note that lag=1 means an index of 0 in python, hence the notation lag-1 in the indexing below.
            The argument is thus referring to the lagged state vector as [1,2,3,4,5,..., nlag]
        """    

        import da.tools.io4 as io
        try:
            import matplotlib.pyplot as plt
        except:
            pass

        # Get the needed matrices from the specified covariance files

        file_ocn_cov = self.DaCycle.DaSystem['ocn.covariance'] 

        cov_files    = os.listdir(self.DaCycle.DaSystem['bio.cov.dir'])

        cov_files    = [os.path.join(self.DaCycle.DaSystem['bio.cov.dir'],f) for f in cov_files if self.DaCycle.DaSystem['bio.cov.prefix'] in f]

        msg = "Found %d covariances to use for biosphere" % len(cov_files) ; logging.debug(msg)

        # replace YYYY.MM in the ocean covariance file string

        file_ocn_cov = file_ocn_cov.replace('2000.01',date.strftime('%Y.%m'))

        cov_files.append(file_ocn_cov)

        covariancematrixlist = []
        for file in cov_files:

            if not os.path.exists(file):

                msg = "Cannot find the specified file %s" % file ; logging.error(msg)
                raise IOError,msg
            else:

                msg = "Using covariance file: %s" % file ; logging.debug(msg)

            f   = io.CT_Read(file,'read')

            if 'pco2' in file: 
                cov_ocn     = f.GetVariable('CORMAT')
                cov         = cov_ocn
            else: 
                cov         = f.GetVariable('covariance')
                #cov_sf      = 10.0/np.sqrt(cov.diagonal().sum())  # this scaling factor makes the total variance close to the value of a single ecoregion
                cov_sf      = 20.0/np.sqrt(cov.diagonal().sum())  # this scaling factor makes the total variance close to the value of a single ecoregion
                cov         = cov*cov_sf

            dummy   = f.close()

            covariancematrixlist.append(cov)

        dummy   = logging.debug("Succesfully closed files after retrieving prior covariance matrices")

        # Once we have the matrices, we can start to make the full covariance matrix, and then decompose it

        return covariancematrixlist

    def MakeNewEnsemble(self,lag, covariancematrixlist = [None]):
        """ 
        :param lag: an integer indicating the time step in the lag order
        :param covariancematrix: a list of matrices specifying the covariance distribution to draw from
        :rtype: None
    
        Make a new ensemble, the attribute lag refers to the position in the state vector. 
        Note that lag=1 means an index of 0 in python, hence the notation lag-1 in the indexing below.
        The argument is thus referring to the lagged state vector as [1,2,3,4,5,..., nlag]

        The covariance list object to be passed holds a list of matrices with a total number of dimensions [nparams, nparams], which is
        used to draw ensemblemembers from. Each draw is done on a matrix from the list, to make the computational burden smaller when
        the StateVector nparams becomes very large.

        """    
        try:
            import matplotlib.pyplot as plt
        except:
            pass

        if not isinstance(covariancematrixlist,list):
            msg = "The covariance matrix or matrices must be passed as a list of array objects, exiting..." ; logging.error(msg)
            raise ValueError

        # Check dimensions of covariance matrix list, must add up to nparams

        dims = 1  # start from 1.0 to account for the last parameter that scales Ice+Non-optimized, we have no covariance matrix for this though

        for matrix in covariancematrixlist: 
            dims += matrix.shape[0]

        if dims != self.nparams:
            msg = "The total dimension of the covariance matrices passed (%d) does not add up to the prescribed nparams (%d), exiting..."%(dims,self.nparams) ; logging.error(msg)
            raise ValueError

        # Loop over list if identity matrices and create a matrix of (nparams,nmembers) with the deviations

        istart      = 0
        istop       = 0
        dof         = 0.0
        dev_matrix  = np.zeros((self.nparams,self.nmembers,),'float')
        randstate    = np.random.get_state()

        for matrix in covariancematrixlist:


            # Make a cholesky decomposition of the covariance matrix

            U,s,Vh  = np.linalg.svd(matrix)
            dof     += np.sum(s)**2/sum(s**2)
            try:
                C       = np.linalg.cholesky(matrix)
            except np.linalg.linalg.LinAlgError,err:
                msg =   'Cholesky decomposition has failed '                    ; logging.error(msg)
                msg =   'For a matrix of dimensions: %d'%matrix.shape[0]        ; logging.error(msg)
                logging.debug(err)
                raise np.linalg.linalg.LinAlgError


            # Draw nmembers instances of this distribution

            npoints      = matrix.shape[0]

            istop        = istop+npoints

            for member in range(1,self.nmembers):
                rands        = np.random.randn(npoints)
                deviations   = np.dot(C,rands)
                dev_matrix[istart:istop,member-1] = deviations
                dev_matrix[istop,member-1] = 1.e-10*np.random.randn()

            #cov2 = np.dot(dev_matrix[istart:istop,:],np.transpose(dev_matrix[istart:istop,:])) / (self.nmembers-1)
            #print matrix.sum(),cov2.sum(),abs(matrix.diagonal()-cov2.diagonal()).max(), matrix.shape,cov2.shape

            istart      = istart + npoints

        msg =   'Successfully constructed a deviation matrix from covariance structure' ; logging.debug(msg)
        msg =   'Appr. degrees of freedom in full covariance matrix is %s'%(int(dof))  ; logging.info(msg)

        # Now fill the ensemble members with the deviations we have just created


        # Create mean values 

        NewMean                     = np.ones(self.nparams,float) # standard value for a new time step is 1.0

        # If this is not the start of the filter, average previous two optimized steps into the mix

        if lag == self.nlag and self.nlag >= 3:
            NewMean                     += self.EnsembleMembers[lag-2][0].ParameterValues + \
                                           self.EnsembleMembers[lag-3][0].ParameterValues 
            NewMean                      = NewMean/3.0

        # Create the first ensemble member with a deviation of 0.0 and add to list

        NewMember                   = self.GetNewMember(0)
        NewMember.ParameterValues   = NewMean.flatten()  # no deviations
        dummy                       = self.EnsembleMembers[lag-1].append(NewMember)

        # Create members 1:nmembers and add to EnsembleMembers list

        for member in range(1,self.nmembers):

            NewMember                   = self.GetNewMember(member)
            NewMember.ParameterValues   = dev_matrix[:,member-1] + NewMean
            dummy                       = self.EnsembleMembers[lag-1].append(NewMember)

        msg =   '%d new ensemble members were added to the state vector # %d'%(self.nmembers,lag)  ; logging.debug(msg)





################### End Class CtStateVector ###################


if __name__ == "__main__":

    import os
    import sys
    from da.tools.initexit import StartLogger , ParseOptions
    from da.tools.pipeline import JobStart
    from da.tools.initexit import CycleControl
    from da.ctgridded.dasystem import CtGriddedDaSystem 
    import numpy as np

    sys.path.append(os.getcwd())
    sys.path.append('../../')

    opts = ['-v']
    args = {'rc':'../../dagridded.rc'}

    StartLogger(level = logging.DEBUG)

    DaCycle     = CycleControl(opts,args)

    DaSystem    = CtGriddedDaSystem('../../da/rc/carbontrackergridded.rc')
    StateVector = CtGriddedStateVector()

    DaSystem.Initialize()
    DaSystem.Validate()
    DaCycle.DaSystem    = DaSystem
    DaCycle.Initialize()
    StateVector.DaCycle = DaCycle # also embed object in StateVector so it can access cycle information for I/O etc

    dummy               = StateVector.Initialize()

    for n in range(1):
        cov   = StateVector.GetCovariance(DaCycle['time.start'])
        dummy = StateVector.MakeNewEnsemble(n+1,cov)

    #StateVector.Propagate()

    #savedir         = './'
    #filename        = os.path.join(savedir,'savestate.nc')
    #dummy = StateVector.WriteToFile(filename)
    #StateVector.ReadFromFile(filename)

    StateVector.StateToTC(fluxvector=np.ones(StateVector.nparams) )

    sys.exit(2)

    members =  StateVector.EnsembleMembers[-1]

    

    for mem in members[0:1]:

        mem.ParameterValues = np.arange(StateVector.nparams)+1.0
        data                = StateVector.VectorToGrid(vectordata=mem.ParameterValues)
        params              = StateVector.VectorToGrid(griddata=data, reverse=True, method = "minval")

        tcparams            = StateVector.VectorToTC(mem.ParameterValues)

    print (StateVector.gridmap-data).min()
    print (StateVector.gridmap-data).max()
    print (mem.ParameterValues-params).min()
    print (mem.ParameterValues-params).max()

