#!/usr/bin/env python
# tm5_tools.py

"""
Author : peters 

Revision History:
File created on 09 Feb 2009.
Major modifications to go to a class-based approach, July 2010.

This module holds specific functions needed to use the TM5 model within the data assimilation shell. It uses the information 
from the DA system in combination with the generic tm5.rc files. 

The TM5 model is now controlled by a python subprocess. This subprocess consists of an MPI wrapper (written in C) that spawns
a large number ( N= nmembers) of TM5 model instances under mpirun, and waits for them all to finish.

The design of the system assumes that the tm5.x (executable) was pre-compiled with the normal TM5 tools, and is residing in a 
directory specified by the ${RUNDIR} of a tm5 rc-file. This tm5 rc-file name is taken from the data assimilation rc-file. Thus,
this python shell does *not* compile the TM5 model for you!

"""

import os, sys
import subprocess
import logging
import shutil

sys.path.append(os.getcwd())
sys.path.append("../../")

identifier      = 'TM5'
version         = 'release 3.0'
mpi_shell_filename  = 'tm5_mpi_wrapper'
mpi_shell_location  = 'da/bin/'


################### Begin Class TM5 ###################

from da.baseclasses.observationoperator import ObservationOperator

class TM5ObservationOperator(ObservationOperator):
    """ This class holds methods and variables that are needed to run the TM5 model. It is initiated with as only argument a TM5 rc-file
        location. This rc-file will be used to figure out the settings for the run. 
        
        *** This method of running TM5 assumes that a pre-compiled tm5.exe is present, and it will be run from time.start to time.final ***
        
        These settings can be modified later. To run a model version, simply compile the model using an existing TM5 rc-file, then 
        open python, and type:

           []> tm=TM5('/Users/peters/Modeling/TM5/tutorial.rc')
           []> tm.WriteRc()
           []> tm.WriteRunRc()
           []> tm.Run()

        To use this class inside a data assimilation cycle, a stand-alone method "Initialize()" is included which modifies the TM5
        settings according to an external dictionary of values to overwrite, and then runs the TM5 model.

    
    """

    def __init__(self,RcFileName, DaCycle=None):
        """ The instance of an TMObservationOperator is application dependent """
        self.Identifier     = self.getid()    # the identifier gives the model name
        self.Version        = self.getversion()       # the model version used
        self.RestartFileList = []
        self.OutputFileList = ()
        self.RcFileType     = 'None'
        self.outputdir      = None # Needed for opening the samples.nc files created 

        dummy               = self.LoadRc(RcFileName)   # load the specified rc-file
        dummy               = self.ValidateRc()         # validate the contents

        # The following code allows the object to be initialized with a DaCycle object already present. Otherwise, it can
        # be added at a later moment.

        if DaCycle != None:
            self.DaCycle = DaCycle
        else:
            self.DaCycle = {}

        msg                 = 'Observation Operator initialized: %s (%s)'%(self.Identifier, self.Version,) ; logging.info(msg)
        
    def Initialize(self ):
        """ 
           Execute all steps needed to prepare the ObsOperator for use inside CTDAS, only done at the very first cycle normally
    
        """
        from da.tools.general import CreateDirs


        if self.DaCycle['time.restart'] == False :

            msg                 = 'First time step, setting up and compiling the TM5 model before proceeding!' ; logging.info(msg)
            # Modify the rc-file to reflect directory structure defined by CTDAS
            NewItems = {
                        'my.basedir'     : self.DaCycle['dir.exec']    ,
                        }

            self.ModifyRC(NewItems)

            # Create the TM5 run directory to hold a copy of the modified rc-file

            tm5compiledir  = self.tm_settings[self.rundirkey]
            dummy          = CreateDirs(tm5compiledir)

            rcfilename      = os.path.join(tm5compiledir,'tm5_setup_init.rc')
            dummy           = self.WriteRc(rcfilename)
            self.RcFileName = rcfilename

            # Compile TM5 in the new directory, but only if this is a fresh start

            self.CompileTM5()

            self.RcFileName = os.path.join(self.tm_settings['rundir'],self.tm_settings['install.rc'])

            # Replace the rc filename for TM5 with the newly created one in the new run directory

            
            msg              = 'Working copy of the tm5.rc file is in place (%s) '%self.RcFileName ; logging.debug(msg)

            # And also replace the path to the ObservationOperator in the DaCycle object so we can work from the TM5 copy from here on

            self.DaCycle['da.obsoperator.rc'] = self.RcFileName

            msg              = '...and set as the da.obsoperator.rc value in this DaCycle '; logging.debug(msg)

            self.LoadRc(self.RcFileName)


        return 0

    def CompileTM5(self):
        """
            Compile TM5 model using setup_tm5 and the modified rc-file
        """
        import subprocess
        from string import join
        DaPlatForm  = self.DaCycle.DaPlatForm

        try:
            os.chdir(self.DaCycle['da.obsoperator.home'])
        except:

            tm5_dir = os.path.split(self.DaCycle['da.obsoperator.rc'])[0]
            msg                 = 'Guessing your TM5 root dir from the rc filename' ; logging.warning(msg)
            msg                 = 'Try adding a key da.obsoperator.home to your da.rc' ; logging.warning(msg)
            msg                 = 'Proceeding from guessed TM5 root dir (%s) '%tm5_dir ; logging.warning(msg)

            os.chdir(tm5_dir)

        cmd = ['python','setup_tm5','--%s'%DaPlatForm.ReturnQueueType(),self.RcFileName]

        msg = 'Starting the external TM5 setup script' ; logging.info(msg)
        msg = 'using command ... %s'% join(cmd) ; logging.info(msg)

        retcode = subprocess.call(cmd)
        os.chdir(self.DaCycle['dir.da_submit'])

        if retcode != 0:
            msg                 = 'Compilation failed, quitting CTDAS' ; logging.error(msg)
            raise IOError
            sys.exit(2)
        else:
            msg                 = 'Compilation successful, continuing' ; logging.info(msg)

    def getid(self):
        return identifier

    def getversion(self):
        return version

    def PrepareRun(self ):
        """ 
        Prepare a forward model TM5 run, this consists of:

          - reading the working copy TM5 rc-file, 
          - validating it, 
          - modifying the values,
          - Removing the existing tm5.ok file if present
    
        """
        from da.tools.general import CreateLinks

        DaPlatForm              = self.DaCycle.DaPlatForm

# First reload the original tm5.rc file to get unmodified settings

        dummy               = self.LoadRc(self.RcFileName)   # load the specified rc-file
        dummy               = self.ValidateRc()         # validate the contents


# Write a modified TM5 model rc-file in which run/break times are defined by our da system

        NewItems = {
                    'submit.options'     : DaPlatForm.ReturnBlockingFlag()    ,
                    self.timestartkey    : self.DaCycle['time.sample.start']    ,
                    self.timefinalkey    : self.DaCycle['time.sample.end']    ,
                    'jobstep.timerange.start'    : self.DaCycle['time.sample.start']    ,
                    'jobstep.timerange.end'      : self.DaCycle['time.sample.end']    ,
                    'ct.params.input.dir' : self.DaCycle['dir.input']    ,
                    'ct.params.input.file' : os.path.join(self.DaCycle['dir.input'],'parameters')    ,
                    'output.flask.infile' : self.DaCycle['ObsOperator.inputfile']
                    }

        if self.DaCycle['time.restart']:  # If this is a restart from a previous cycle, the TM5 model should do a restart
            NewItems[self.istartkey] = self.restartvalue
            msg                 = 'Resetting TM5 to perform restart' ; logging.debug(msg)
        else:
            NewItems[self.istartkey] = self.coldstartvalue  # if not, start TM5 'cold'
            msg                 = 'Resetting TM5 to perform cold start' ; logging.debug(msg)

        if self.DaCycle['time.sample.window'] != 0:  # If this is a restart from a previous time step within the filter lag, the TM5 model should do a restart
            NewItems[self.istartkey] = self.restartvalue
            msg                 = 'Resetting TM5 to perform restart' ; logging.debug(msg)
        
        # If neither one is true, simply take the istart value from the tm5.rc file that was read

        self.ModifyRC(NewItems)

        dummy = self.WriteRc(self.RcFileName)

        return 0

    def LoadRc(self,RcFileName):
        """ 
        This method loads a TM5 rc-file with settings for this simulation 
        """
        import da.tools.rcn as rc

        self.rcfile         = rc.RcFile(RcFileName)
        self.tm_settings    = self.rcfile.values
        self.RcFileName     = RcFileName
        self.Tm5RcLoaded    = True

        if 'my.source.dirs' in self.tm_settings.keys():
            self.RcFileType = 'pycasso'
        else:
            self.RcfIleType = 'pre-pycasso'

        msg                 = 'TM5 rc-file loaded successfully' ; logging.debug(msg)

        return True

    def ValidateRc(self):
        """
        Validate the contents of the tm_settings dictionary and add extra values. The required items for the TM5 rc-file
        are specified in the tm5_tools module, as dictionary variable "needed_rc_items".
        """
        from da.tools.general import ToDatetime

        if self.RcFileType == 'pycasso':

            self.projectkey      = 'my.project.dir'
            self.rundirkey       = 'my.run.dir'
            self.outputdirkey    = 'output.dir'
            self.savedirkey      = 'restart.write.dir'
            self.timestartkey    = 'timerange.start'
            self.timefinalkey    = 'timerange.end'
            self.timelengthkey   = 'jobstep.length'
            self.istartkey       = 'istart'
            self.restartvalue    = 33
            self.coldstartvalue  = 9

        else:

            self.projectkey      = 'runid'
            self.rundirkey       = 'rundir'
            self.outputdirkey    = 'outputdir'
            self.savedirkey      = 'savedir'
            self.timestartkey    = 'time.start'
            self.timefinalkey    = 'time.final'
            self.timelengthkey   = 'time.break.nday'
            self.istartkey       = 'istart'
            self.restartvalue    = 3
            self.coldstartvalue  = 9

        needed_rc_items = [
                            self.projectkey    ,
                            self.rundirkey    ,
                            self.outputdirkey ,
                            self.savedirkey   ,
                            self.timestartkey ,
                            self.timefinalkey ,
                            self.timelengthkey,
                            self.istartkey
                          ]

        for k,v in self.tm_settings.iteritems():
            if v == 'True' : self.tm_settings[k] = True
            if v == 'False': self.tm_settings[k] = False
            if 'date' in k : self.tm_settings[k] = ToDatetime(v)
            if 'time.start' in k : 
                self.tm_settings[k]              = ToDatetime(v,fmt='TM5')
            if 'time.final' in k : 
                self.tm_settings[k]              = ToDatetime(v,fmt='TM5')
            if 'timerange.start' in k : 
                self.tm_settings[k]              = ToDatetime(v)
            if 'timerange.end' in k : 
                self.tm_settings[k]              = ToDatetime(v)

        for key in needed_rc_items:

            if not self.tm_settings.has_key(key):
                status,msg = ( False,'Missing a required value in rc-file : %s' % key)
                logging.error(msg)
                raise IOError,msg

        status,msg = ( True,'rc-file has been validated succesfully' )  ; logging.debug(msg)




    def ModifyRC(self,NewValues):
        """ 
        Modify parts of the tm5 settings, for instance to give control of file locations to the DA shell
        instead of to the tm5.rc script. 

        Note that we replace these values in all {key,value} pairs of the tm5.rc file!

        """
    
        for k,v in NewValues.iteritems():
            if self.tm_settings.has_key(k):
                # keep previous value
                v_orig= self.tm_settings[k]
                #replace with new
                self.tm_settings[k] = v
                #replace all instances of old with new, but only if it concerns a name of a path!!!
                if os.path.exists(str(v)): 
                    for k_old,v_old in self.tm_settings.iteritems():
                        if not isinstance(v_old,str): 
                            continue
                        if str(v_orig) in str(v_old): 
                            v_new = str(v_old).replace(str(v_orig),str(v))
                            self.tm_settings[k_old] = v_new

                logging.debug('Replaced tm5 rc-item %s ' % k)

            else:
                self.tm_settings[k] = v
                logging.debug('Added new tm5 rc-item %s ' % k)


    def WriteRc(self,tm5rcfilename):
        """ 
        Write the rc-file settings to a tm5.rc file in the rundir
        """
        import da.tools.rc as rc

        dummy           = rc.write(tm5rcfilename,self.tm_settings)

        msg             = "Modified rc file for TM5 written (%s)" % tm5rcfilename      ;logging.debug(msg)

        return None

    def WriteRunRc(self):
        """ 
        Create the tm5-runtime.rc file which is read by initexit.F90 to control time loop and restart from save files
        """
        import da.tools.rc as rc

        tm5rcfilename   = os.path.join(self.tm_settings[self.rundirkey],'tm5_runtime.rc')

        dummy           = rc.write(tm5rcfilename,self.tm_settings)

        rc_runtm5={}

        rc_runtm5['year1']  = self.tm_settings[self.timestartkey].year
        rc_runtm5['month1'] = self.tm_settings[self.timestartkey].month
        rc_runtm5['day1']   = self.tm_settings[self.timestartkey].day
        rc_runtm5['hour1']  = self.tm_settings[self.timestartkey].hour
        rc_runtm5['minu1']  = 0
        rc_runtm5['sec1']   = 0

        rc_runtm5['year2']  = self.tm_settings[self.timefinalkey].year
        rc_runtm5['month2'] = self.tm_settings[self.timefinalkey].month
        rc_runtm5['day2']   = self.tm_settings[self.timefinalkey].day
        rc_runtm5['hour2']  = self.tm_settings[self.timefinalkey].hour
        rc_runtm5['minu2']  = 0
        rc_runtm5['sec2']   = 0

        rc_runtm5[self.istartkey]    = self.tm_settings[self.istartkey]
        rc_runtm5[self.savedirkey]   = self.tm_settings[self.savedirkey]
        rc_runtm5[self.outputdirkey] = self.tm_settings[self.outputdirkey]

        rc.write(tm5rcfilename,rc_runtm5)

        msg             = "Modified tm5_runtime.rc written (%s)" % tm5rcfilename      ;logging.debug(msg)


    def ValidateInput(self):
        """
        Make sure that parameter files are written to the TM5 inputdir, and that observation lists are present
        """

        datadir   = self.tm_settings['ct.params.input.dir']
        if not os.path.exists(datadir):
            msg   = "The specified input directory for the TM5 model to read from does not exist (%s), exiting..."%datadir ; logging.error(msg)
            raise IOError,msg

        datafiles = os.listdir(datadir)

        obsfile   = self.DaCycle['ObsOperator.inputfile']

        if not os.path.exists(obsfile):
            msg   = "The specified obs input file for the TM5 model to read from does not exist (%s), exiting..."%obsfile ; logging.error(msg)
            raise IOError,msg


        for n in range(int(self.DaCycle['da.optimizer.nmembers'])):
            paramfile = 'parameters.%03d.nc'%n
            if paramfile not in datafiles:
                msg = "The specified parameter input file for the TM5 model to read from does not exist (%s), exiting..."%paramfile ; logging.error(msg)
                raise IOError,msg

        # Next, make sure there is an actual model version compiled and ready to execute

        targetdir          = os.path.join(self.tm_settings[self.rundirkey])

        if self.RcFileType == 'pycasso':
            self.Tm5Executable = os.path.join(targetdir,self.tm_settings['my.basename']+'.x')
        else:
            self.Tm5Executable = os.path.join(targetdir,'tm5.x')

        if not os.path.exists(self.Tm5Executable):
            msg = "Required TM5 executable was not found %s"%self.Tm5Executable                           ; logging.error(msg)
            msg = "Please compile the model with the specified rc-file and the regular TM5 scripts first" ; logging.error(msg)
            raise IOError

        return 0

    def GetInitialData(self):
        """ This method places all initial data needed by an ObservationOperator in the proper folder for the model.
            For TM5, this means copying the save_*.hdf* files to the dir.save directory from which TM5 will read initial
            concentrations for all tracers. 

            We get the input data from the restart.current directory at 2 times:
                (1) When the model starts the forecast over nlag cycles
                (2) When the model starts the advance step over 1 cycle


         """

        msg     = "Moving TM5 model restart data from the restart/current directory to the TM5 save dir" ; logging.debug(msg)

        # First get the restart data for TM5 from the current restart dir of the filter

        sourcedir   = self.DaCycle['dir.restart.current']
        targetdir   = self.tm_settings[self.savedirkey]
        self.outputdir      = self.tm_settings[self.outputdirkey]  #IV test

        for file in os.listdir(sourcedir):

            file = os.path.join(sourcedir,file)

            if os.path.isdir(file): # skip dirs

                msg         = "           [skip] .... %s " % file                                     ; logging.debug(msg)
                continue    
            #if not file.startswith('save_'):
            if not file.startswith('TM5_restart'):

                msg         = "           [skip] .... %s " % file                                     ; logging.debug(msg)
                continue    

            # all okay, copy file

            msg         = "           [copy] .... %s " % file                                                ; logging.debug(msg)
            dummy       = shutil.copy(file,file.replace(sourcedir,targetdir) )

        msg     = "All restart data have been copied from the restart/current directory to the TM5 save dir" ; logging.debug(msg)


        return None


    def Run(self):
        """ 
         Start the TM5 executable. A new log file is started for the TM5 model IO, and then a subprocess is
         spawned with the tm5_mpi_wrapper and the tm5.x executable. The exit code of the model is caught and
         only if successfull on all processors will execution of the shell continue. 
         
        """

        cwd = os.getcwd()

        # From here on, several options should be implemented.

        #
        # (1) Where an mpi process is forked to do a TM5 instance with N tracers, each an ensemble member
        #
        # (2) Where N processes are spawned, each being one TM5 instance representing one member
        #
        # (3) Where N/m processes are spawned, each being a TM5 instance that handles m ensemble members
        #
        # In principle, it is best to make these processes produce scripts that can be executed stand-alone, or
        # be submitted to a queue.
        #

        # Open logfile and spawn model, wait for finish and return code

        # Code for Option (1)

        code = self.TM5_With_N_tracers()

        if code == 0:
            logging.info('Finished model executable succesfully (%s)'%code)
            self.Status = 'Success'
        else:
            logging.error('Error in model executable return code: %s ' % code)
            self.Status = 'Failed'
            raise OSError

        # Return to working directory

        os.chdir(cwd)

        return code

    def TM5_under_mpirun(self ):
        """ Method handles the case where a shell runs an MPI process that forks into N TM5 model instances """
        from string import join
        import datetime

        DaPlatForm              = self.DaCycle.DaPlatForm

        targetdir               = os.path.join(self.tm_settings[self.rundirkey])

        if not os.path.exists(os.path.join(mpi_shell_location,mpi_shell_filename) ):
            msg = "Cannot find the mpi_shell wrapper needed for completion (%s) in (%s)"% (mpi_shell_filename,mpi_shell_location) ; logging.error(msg)
            msg = "Please see the %s/readme_wrapper.txt file for instructions to compile it"%mpi_shell_location ; logging.error(msg)
            raise IOError

        shutil.copy(os.path.join(mpi_shell_location,mpi_shell_filename) ,os.path.join(targetdir,mpi_shell_filename) )

        # Go to executable directory and start the subprocess, using a new logfile

        dummy                   = os.chdir(targetdir)
        msg                     = 'Changing directory to %s ' % targetdir       ;logging.debug(msg)

        # Remove the tm5.ok file from a previous run, placed back only if a successful TM5 run is executed

        okfile                  = 'tm5.ok'
        if os.path.exists(okfile): 
            dummy               = os.remove(okfile)

        nthreads                = self.DaCycle['da.optimizer.nmembers']
        jobparams               = {'jobname':'tm5',
                                   'jobnodes':'ncomp %d'%int(nthreads),
                                   'jobtime':'00:30:00',
                                   'joblog':os.path.join(self.DaCycle['dir.jobs'])
                                  }

        # file ID and names
        jobid                   = 'tm5'
        targetdir               = os.path.join(self.DaCycle['dir.exec'])
        jobfile                 = os.path.join(targetdir,'jb.%s.jb'%jobid)
        logfile                 = jobfile.replace('.jb','.log')

        template                = DaPlatForm.GetJobTemplate(jobparams,block=True)
        template                += 'cd %s\n'%targetdir
        template                += 'mpirun -np %d %s ./tm5.x\n'%(int(nthreads),mpi_shell_filename,)

        dummy                   = DaPlatForm.WriteJob(jobfile,template,jobid)
      
        msg                     = 'Submitting job at %s'%datetime.datetime.now() ; logging.info(msg)
        code                    = DaPlatForm.SubmitJob(jobfile,joblog=jobfile) 
        msg                     = 'Resuming job at %s'%datetime.datetime.now() ; logging.info(msg)
        
        if not os.path.exists(okfile): 
            code = -1
        else:
            code = 0

        return code


    def TM5_With_N_tracers(self):
        """ Method handles the case where one TM5 model instance with N tracers does the sampling of all ensemble members"""
        from string import join
        import datetime
        import subprocess
        import os
        DaPlatForm              = self.DaCycle.DaPlatForm

        tm5submitdir               = os.path.join(self.tm_settings[self.rundirkey])
        msg                     = 'tm5submitdir',tm5submitdir ; logging.info(msg)

        # Go to executable directory and start the subprocess, using a new logfile

        dummy                   = os.chdir(tm5submitdir)
        msg                     = 'Changing directory to %s ' % tm5submitdir       ;logging.debug(msg)

        # Remove the tm5.ok file from a previous run, placed back only if a successful TM5 run is executed

        okfile                  = 'tm5.ok'
        if os.path.exists(okfile): 
            dummy               = os.remove(okfile)

        # Prepare a job for the current platform, this job needs to account for all job parameters such as
        # runtime, queue request, number of processors, and other platform specific parameters

        # It is easiest if all of these options are 'pre-configured' through the tm5.rc file that is used, then
        # running the ObsOperator on all platforms simply becomes a matter of running the ./setup_tm5 script
        # that was also used to compile TM5 when not actually running CTDAS. The first run of CTDAS
        # then could be one where TM5 is actually compiled and run, the rest of the jobs would then *not* re-compile.
        # An option would need to be added to force a re-compile of the TM5 code, for debugging purposes.

        # file ID and names
        submitcommand           = self.tm_settings['submit.command'] 
        msg                     = 'Submitting job at %s'%datetime.datetime.now() ; logging.info(msg)
        code                    = subprocess.call(submitcommand.split()) 
        msg                     = 'Resuming job at %s'%datetime.datetime.now() ; logging.info(msg)
        
        if not os.path.exists(okfile): 
            code = -1
        else:
            code = 0

        return code

    def SaveData(self):
        """ Copy the TM5 recovery data from the outputdir to the TM5 savedir, also add the restart files to a list of names
            that is used by the DaCycle object to collect restart data for the filter.

            WP Note: with the new pycasso restart files we no longer need to copy save files from outdir to savedir

            Note 2: also adding the weekly mean flux output to the OutputFileList for later collection
         """

        sourcedir   = os.path.join(self.tm_settings[self.savedirkey])
        filter      = ['%s'%self.tm_settings[self.timefinalkey].strftime('%Y%m%d')]

        msg         = "Creating a new list of TM5 restart data"                                 ; logging.debug(msg)
        msg         = "           from directory: %s " % sourcedir                                ; logging.debug(msg)
        msg         = "           with filter: %s " % filter                                      ; logging.debug(msg)


        # Start from empty lists for each TM5 run. Note that these "private" lists from the obs operator are later on appended to the system
        # lists

        self.RestartFileList = []

        for file in os.listdir(sourcedir):

            file = os.path.join(sourcedir,file)

            if os.path.isdir(file): # skip dirs
                skip = True
            elif filter == []:      # copy all
                skip= False        
            else:                   # check filter
                skip = True         # default skip
                for f in filter:
                    if f in file: 
                        skip = False # unless in filter
                        break
                
            if skip: 
                msg         = "           [skip] .... %s " % file                                     ; logging.debug(msg)
                continue    

            dummy       = self.RestartFileList.append( file   )
            msg         = "           [added to restart list] .... %s " % file                               ; logging.debug(msg)

        sourcedir   = os.path.join(self.tm_settings[self.outputdirkey])
        sd_ed      = self.DaCycle['time.sample.stamp']
        filter      = ['flask_output.%s'%sd_ed,'flux1x1_%s'%sd_ed]

        msg         = "Creating a new list of TM5 output data to collect"                         ; logging.debug(msg)
        msg         = "           from directory: %s " % sourcedir                                ; logging.debug(msg)
        msg         = "           with filter: %s " % filter                                      ; logging.debug(msg)


        # Start from empty lists for each TM5 run. Note that these "private" lists from the obs operator are later on appended to the system
        # lists

        self.OutputFileList  = []

        for file in os.listdir(sourcedir):

            file = os.path.join(sourcedir,file)

            if os.path.isdir(file): # skip dirs
                skip = True
            elif filter == []:      # copy all
                skip= False        
            else:                   # check filter
                skip = True         # default skip
                for f in filter:
                    if f in file: 
                        skip = False # unless in filter
                        break
                
            if skip: 
                msg         = "           [skip] .... %s " % file                                     ; logging.debug(msg)
                continue    

            dummy       = self.OutputFileList.append( file  )
            msg         = "           [added to output list] .... %s " % file                               ; logging.debug(msg)


################### End Class TM5 ###################
                   

if __name__ == "__main__":

    sys.path.append('../../')

    from da.tools.initexit import StartLogger
    from da.tools.pipeline import JobStart
    import datetime as dtm
    import sys,os

    sys.path.append(os.getcwd())

    StartLogger()

    DaCycle = JobStart([],{'rc':'da.rc'})
    DaCycle.Initialize()
    DaCycle['time.sample.start'] = dtm.datetime(2000,1,1)
    DaCycle['time.sample.end'] = dtm.datetime(2000,1,2)
    DaCycle['time.sample.window'] = 0

    tm=TM5ObservationOperator()
    tm.Setup()
    tm.Initialize()
    tm.Run()
    tm.SaveData()





