#!/usr/bin/env python
# tools_da.py

"""
Author : peters 

Revision History:
File created on 03 Oct 2008.

Temporary module to hold classes and methods that are in development

"""

import logging
import os
import sys
import shutil
import datetime
import da.tools.rc as rc

def ValidateRC(rcfile,needed_items):
    """ validate the contents of an rc-file given a dictionary of required keys """

    for k,v in rcfile.iteritems():
        if v == 'True' : rcfile[k] = True
        if v == 'False': rcfile[k] = False
        if 'date' in k : rcfile[k] = datetime.datetime.strptime(v,'%Y-%m-%d %H:%M:%S')

    for key in needed_items:

        if not rcfile.has_key(key):
            status,msg = ( False,'Missing a required value in rc-file : %s' % key)
            logging.error(msg)
            raise IOError,msg

    status,msg = ( True,'rc-file has been validated succesfully' )  ; logging.debug(msg)


def CreateDirs(dirname,forceclean=False):
    """ Create a directory and report success, only if non-existent """

    if forceclean:
        try:
            shutil.rmtree(dirname)
        except:
            pass

    if not os.path.exists(dirname):
        os.makedirs(dirname)
        msg='Creating new directory %s' % dirname
        logging.info(msg)
    else:
        msg='Using existing directory %s' % dirname
        logging.debug(msg)
    return dirname

def CreateLinks(sourcedir,targetdir):
    """ Create a symbolic link from, source to target and report success, note
        that exisiting links are first removed and then recreated """

    if os.path.exists(targetdir):
        os.unlink(targetdir)
        msg='Unlinking existing directory %s' % targetdir
        logging.debug(msg)
    try:
        os.symlink(sourcedir,targetdir)
        msg='Created new link from %s to %s' % (sourcedir, targetdir)
        logging.debug(msg)
    except OSError,msg:
        msg='Failed to create link from %s to %s' % (sourcedir, targetdir)
        logging.error(msg)
        raise OSError

    return None



def AdvanceTime(time_in,interval):
    """ Advance time_in by a specified interval"""

    time_out=time_in

    if interval == 'month':                       # if monthly, this run will go to the first day of the next month
        if time_in.month != 12: 
            time_out = datetime.datetime(time_in.year,time_in.month+1,1,time_in.hour,0,0)
        else: 
            time_out = datetime.datetime(time_in.year+1,1,1,time_in.hour,0,0)  # end of year provision
    elif interval == 'week':
        time_out = time_in + datetime.timedelta(days=7)
    elif isinstance(interval,datetime.timedelta):
        time_out = time_in + interval
    else:                    # assume that the interval specified is the number of days to run forward before resubmitting
        time_out = time_in + datetime.timedelta(days=float(interval))

    return time_out



def ToDatetime(datestring,fmt=None):
    """ convert a date string to a datetime object """
    import datetime
  
    if fmt == 'TM5':
        datestring = '%04s-%02s-%02s %02s:%02s:00'%(datestring[0:4],datestring[4:6],datestring[6:8],datestring[8:10],datestring[10:12])
    elif fmt == 'pycasso-TM5':
        pass # Format already compatible
    else:
        pass 


    try:
        return datetime.datetime.strptime(datestring,'%Y-%m-%d %H:%M:%S')
    except:
        date,time = datestring.split(' ')
        year,month,day = map(int,date.split('-'))
        hour,minute,second = map(int,time.split(':'))
        return datetime.datetime(year,month,day,hour,minute,second)

def ToDectime(dd):
    """ convert a datetime object to a decimal date """
    import datetime
    import calendar
    from matplotlib.pylab import date2num

    Days0=date2num(datetime.datetime(dd.year,1,1))
    if calendar.isleap(dd.year):
        DaysPerYear=366.
    else:
        DaysPerYear=365.
    DayFrac=date2num(dd)
    return dd.year+(DayFrac-Days0)/DaysPerYear

def StoreData(pickleobject,filename):
    """ pickle object into a specified file name """
    import cPickle

    f = open(filename,'wb')
    dummy = cPickle.dump(pickleobject,f,-1)
    f.close()

    return None

def RestoreData(filename):
    """ unpickle object into a specified file name """
    import cPickle

    f = open(filename,'rb')
    pickleobject = cPickle.load(f)
    f.close()

    return pickleobject


def NameConvert( name=None, to=None ):
   """ Convert between old GLOBALVIEW-style and new ObsPack-style

      print NameConvert(name="lef_surface-pfp_1", to='GV' )
            lef_01P0

      print NameConvert(name="hun_35C3", to='OP' )
            hun_tower-insitu_35
   """

   import re

   identifier = 'NameConvert'

   if name == None or to == None: return ""

   platform_dict = { 'surface':0, 'shipboard':1, 'aircraft':2, 'tower':3 }
   strategy_dict = { 'flask':'D', 'pfp':'P', 'insitu':'C' }

   #----------------------------------->>>
   if to.upper() == "GV":

      # alt_surface-flask_1 -> alt_01D0

      try:
         fields = name.split( '_' )
         code = fields[0]
         lab_num = int( fields[2] )
      except:
         return ""

      try:
         fields = fields[1].split( '-' )
         platform = fields[0]
         strategy = fields[1]
      except:
         return ""

      platform_num = [ v for k,v in platform_dict.iteritems() if platform.lower() == k ]
      if len( platform_num ) == 0:
         print '%s: Platform %s not found in platform dict.' % ( identifier, platform )
         return ""

      strategy_char = [ v for k,v in strategy_dict.iteritems() if strategy.lower() == k ]
      if len( strategy_char ) == 0:
         print '%s: Strategy %s not found in strategy dict.' % ( identifier, strategy )
         return "" 

      return "%s_%2.2d%1s%1d" % ( code, lab_num, strategy_char[0].upper(), int( platform_num[0] ) )

   #----------------------------------->>>
   if to.upper() == "OP":

      # hun_35C3 -> hun_tower-insitu_35

      try:
         fields = name.split( '_' )

         code = fields[0]
         lab_num = int( fields[1][:2] )
         strategy_char = fields[1][2]
         platform_num = int( fields[1][3] )
      except:
         return ""

      platform = [ k for k,v in platform_dict.iteritems() if v == platform_num ]
      if len( platform ) == 0:
         print '%s: Platform number %s not found in platform dict.' % ( identifier, platform_num )
         return ""

      pattern = re.compile( strategy_char, re.IGNORECASE )
      strategy = [ k for k,v in strategy_dict.iteritems() if pattern.search(v) ]
      if len( strategy ) == 0:
         print '%s: Strategy character %s not found in strategy list.' % ( identifier, strategy_char )
         return ""

      return "%s_%s-%s_%d" % ( code, platform[0], strategy[0], lab_num )

if __name__ == "__main__":
    pass
