#!/usr/bin/env python
# da_initexit.py

"""
.. module:: initexit
.. moduleauthor:: Wouter Peters 

Revision History:
File created on 13 May 2009.

The CycleControl class is found in the module :mod:`initexit`. It is derived from the standard python :class:`dictionary` object. It is the only core object of CTDAS that is automatically created in the pipeline, the user (normally) does not need to modify or extend it. The class is created based on options and arguments passes on the command line when submitting your main CTDAS job. 

Valid options are defined in 

.. autofunction:: da.tools.initexit.ParseOptions

With the name of a valid ``rc-file``, the CycleControl object is instantiated and validated. An example rc-file looks
like this:::

    ! Info on the data assimilation cycle

    time.restart        : False                     ! Restart from an existing run T/F
    time.start          : 2000-01-01 00:00:00       ! Start time of first cycle
    time.finish         : 2000-01-08 00:00:00       ! End time of last cycle
    time.cycle          : 7                         ! length of each cycle, 7 means one week
    time.nlag           : 5                         ! number of cycles in one smoother window
    dir.da_run          : ${HOME}/tmp/test_da       ! the run directory for you project

    ! Info on the DA system used

    da.system           : CarbonTracker             ! an identifier for your inversion system
    da.system.rc        : da/rc/carbontracker.rc    ! the settings needed in your inversion system

    ! Info on the forward model to be used

    da.obsoperator         : TM5                                ! an identifier for your observation operator
    da.obsoperator.rc      : ${HOME}/Modeling/TM5/tm5-ctdas.rc  ! the rc-file needed to run youobservation operator
    da.optimizer.nmembers  : 30                                 ! the number of ensemble members desired in the optimization

The most important method of the CycleControl object are listed below:

.. autoclass:: da.tools.initexit.CycleControl 
   :members: Initialize, Finalize,  CollectRestartData, MoveRestartData, 
             SubmitNextCycle, CleanUpCycle, SetupFileStructure, RecoverRun, RandomSeed

Two important attributes of the CycleControl object are:
    (1) DaSystem, an instance of a :ref:`dasystem`
    (2) DaPlatForm, an instance of a :ref:`platform`

Other functions in the module initexit that are related to the control of a DA cycle are:

.. autofunction:: da.tools.initexit.StartLogger 
.. autofunction:: da.tools.initexit.ValidateOptsArgs 


"""

needed_da_items=[
    'time.start',
    'time.finish',
    'time.nlag',
    'time.cycle',
    'dir.da_run',
    'da.system',
    'da.system.rc',
    'da.obsoperator',
    'da.obsoperator.rc',
    'da.optimizer.nmembers']

# only needed in an earlier implemented where each substep was a separate job
# validprocesses = ['start','done','samplestate','advance','invert']

import logging
import os
import sys
import shutil
import datetime



class CycleControl(dict):
    """
    This object controls the CTDAS system flow and functionality.
    """
        
    def __init__(self,opts=[],args={}):
        """
        The CycleControl object is instantiated with a set of options and arguments.
        The list of arguments must contain the name of an existing ``rc-file``. 
        This rc-file is loaded by method :meth:`~da.tools.initexit.CycleControl.LoadRc` and validated
        by :meth:`~da.tools.initexit.CycleControl.ValidateRC`

        Options for the CycleControl consist of accepted command line flags or arguments 
        in :func:`~da.tools.initexit.CycleControl.ParseOptions`

        """

        self.LoadRc(args['rc'])
        self.ValidateRC()
        self.opts = opts

        # Add some useful variables to the rc-file dictionary

        self['jobrcfilename']    = self.RcFileName
        self['dir.da_submit']    = os.getcwd()
        self['da.crash.recover'] = '-r' in opts
        self['verbose']          = '-v' in opts
        self.DaSystem            = None # to be filled later
        self.RestartFileList     = [] # List of files needed for restart, to be extended later
        self.OutputFileList      = [] # List of files needed for output, to be extended later

    def __str__(self):
        """
        String representation of a CycleControl object
        """

        msg = "==============================================================="    ; print msg
        msg = "DA Cycle rc-file is %s" % self.RcFileName                                ; print msg
        msg = "DA Cycle run directory is %s" % self['dir.da_run']           ; print msg
        msg = "DA Cycle inverse system is %s" % self['da.system']           ; print msg
        msg = "DA Cycle obs operator is %s" % self['da.obsoperator']        ; print msg
        msg = "==============================================================="    ; print msg

        return ""


    def LoadRc(self,RcFileName):
        """ 
        This method loads a DA Cycle rc-file with settings for this simulation 
        """
        import da.tools.rc as rc

        rcdata = rc.read(RcFileName)
        for k,v in rcdata.iteritems():
            self[k] = v
        self.RcFileName     = RcFileName
        self.DaRcLoaded    = True

        msg                 = 'DA Cycle rc-file (%s) loaded successfully'%self.RcFileName ; logging.info(msg)

        return True


    def ValidateRC(self):
        """ 
        Validate the contents of the rc-file given a dictionary of required keys. 
        Currently required keys are :attr:`~da.tools.initexit.needed_da_items`
        """
        from da.tools.general import ToDatetime

        for k,v in self.iteritems():
	    if v in ['True','true', 't','T', 'y', 'yes']:
		self[k] = True
	    if v in ['False','false', 'f','F', 'n', 'no']:
		self[k] = False
            if 'date' in k : self[k] = ToDatetime(v)
            if 'time.start' in k : 
                self[k] = ToDatetime(v)
            if 'time.end' in k : 
                self[k] = ToDatetime(v)
            if 'time.finish' in k : 
                self[k] = ToDatetime(v)

        for key in needed_da_items:

            if not self.has_key(key):
                status,msge = ( False,'Missing a required value in rc-file : %s' % key)  
                logging.error(msge)
                msg = '!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! ' ; logging.error(msg)
                msg = 'Please note the update on Dec 02 2011 where rc-file names for DaSystem and ' ; logging.error(msg)
                msg = 'are from now on specified in the main rc-file (see da/rc/da.rc for example)' ; logging.error(msg)
                msg = '!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! ' ; logging.error(msg)
                raise IOError,msge

        status,msg = ( True,'DA Cycle settings have been validated succesfully' )  ; logging.debug(msg)

        return None

    def ParseTimes(self):
        """ 
        Parse time related parameters into datetime objects for later use 
        """
        from da.tools.general import AdvanceTime

        startdate = self['time.start']
        finaldate = self['time.finish']                  

        if finaldate <= startdate:
            msg   = 'The start date (%s) is not greater than the end date (%s), please revise'%(startdate.strftime('%Y%m%d'),finaldate.strftime('%Y%m%d')) 
            logging.error(msg)
            raise ValueError
        #
        cyclelength = self['time.cycle']                 # get time step

# Determine end date

        if cyclelength == 'infinite':
            enddate = finaldate
        else:
            enddate = AdvanceTime(startdate,cyclelength)

        dt      = enddate-startdate

        #
        if enddate > finaldate:  # do not run beyon finaldate
            enddate = finaldate

        self['time.start']   = startdate
        self['time.end']     = enddate
        self['time.finish']  = finaldate
        self['cyclelength']  = dt

        msg = "==============================================================="    ; logging.info(msg)
        msg = "DA Cycle start date is %s"   % startdate.strftime('%Y-%m-%d %H:%M')      ; logging.info(msg)
        msg = "DA Cycle end   date is %s"   % enddate.strftime('%Y-%m-%d %H:%M')        ; logging.info(msg)
        msg = "DA Cycle final date is %s"   % finaldate.strftime('%Y-%m-%d %H:%M')      ; logging.info(msg)
        msg = "DA Cycle cycle length is %s" % cyclelength                               ; logging.info(msg)
        msg = "DA Cycle restart is %s"      % str(self['time.restart'])     ; logging.info(msg)
        msg = "==============================================================="    ; logging.info(msg)

        return None

    def SetSampleTimes(self,lag):
        """
        Set the times over which a sampling interval will loop, depending on 
        the lag. Note that lag falls in the interval [0,nlag-1]
        """
        import copy

        # Start from cycle times 
        self['time.sample.start']   = copy.deepcopy(self['time.start'])
        self['time.sample.end']     = copy.deepcopy(self['time.end'])

        # Now advance depending on lag

        for l in range(lag):
            dummy  = self.AdvanceSampleTimes()

        return None

    def AdvanceSampleTimes(self):
        """ 
        Advance sampling start and end time by one cycle interval
        """
        from da.tools.general import AdvanceTime

        startdate = self['time.sample.start']
        enddate   = self['time.sample.end']                  
        cyclelength   = self['cyclelength']                  

        startdate               = AdvanceTime(startdate,cyclelength.days)
        enddate                 = AdvanceTime(enddate,cyclelength.days)

        self['time.sample.start']   = startdate
        self['time.sample.end']     = enddate
    
        return None

    def AdvanceCycleTimes(self):
        """ 
        Advance cycle start and end time by one cycle interval
        """
        from da.tools.general import AdvanceTime

        startdate       = self['time.start']
        enddate         = self['time.end']                  
        cyclelength     = self['cyclelength']                  

        startdate       = AdvanceTime(startdate,cyclelength.days)
        enddate         = AdvanceTime(enddate,cyclelength.days)

        filtertime              = startdate.strftime('%Y%m%d')
        self['dir.output']      = os.path.join(self['dir.da_run'],'output',filtertime)

        self['time.start']   = startdate
        self['time.end']     = enddate

        return None

    def RandomSeed(self,action='read'):
        """ 
        Get the randomseed and save it, or read the random seed and set it. The seed is currently stored
        in a python :mod:`pickle` file, residing in the ``exec`` directory

        """
        import cPickle
        import numpy as np

        filename = os.path.join(self['dir.exec'],'randomseed.pickle')

        if action == 'write':
            f        = open(filename,'wb')
            seed     = np.random.get_state()
            dummy    = cPickle.dump (seed,f,-1)
            dummy    = f.close()

            msg      = "Saved the random seed generator values to file"

        if action == 'read':
            f        = open(filename,'rb')
            seed     = cPickle.load(f)
            dummy    = np.random.set_state(seed)
            dummy    = f.close()

            msg      = "Retrieved the random seed generator values from file"

        logging.info(msg)

        dummy    = self.RestartFileList.extend([filename])

        msg      = "Added the randomseed.pickle file to the RestartFileList" ; logging.debug(msg)


        return None

    def Initialize(self):
        """ 
        This method determines how to proceed with the cycle. Three options are implemented:

            1. *Fresh start*  : set up the required file structure for this simulation and start
            2. *Restart*      : use latest da_runtime variables from the exec dir and restart
            3. *Recover*      : restart after crash by getting data from restart/one-ago folder

        The choice that gets executed depends on the presence of 

            # the ``-r`` option on the command line, this triggers a recover
            # the ``time.restart : True`` option in the da.rc file

        The latter is automatically set if the filter submits the next cycle at the end of the current one, 
        through method :meth:`~da.tools.initexit.CycleControl.SubmitNextCycle`.

        The specific call tree under each scenario is: 

            1. *Fresh Start*
                * dummy = :meth:`~da.tools.initexit.CycleControl.SetupFileStructure()`  <- Create directory tree
            2. *Restart*
                * dummy = :meth:`~da.tools.initexit.CycleControl.SetupFileStructure()`
                * dummy = :meth:`~da.tools.initexit.CycleControl.RandomSeed`    <- Read the random seed from file
            3. *Recover*
                * dummy = :meth:`~da.tools.initexit.CycleControl.SetupFileStructure()`
                * dummy = :meth:`~da.tools.initexit.CycleControl.RecoverRun()`          <- Recover files from restart/one-ago dir, reset ``time.start``
                * dummy = :meth:`~da.tools.initexit.CycleControl.RandomSeed` 

        And is always followed by a call to

            * ParseTimes()
            * WriteRc('jobfilename')


        """        

#
# case 1: A recover from a previous crash, this is signaled by flag "-r"
#
        if self['da.crash.recover']:

            msg   = "Recovering simulation from data in: %s" % self['dir.da_run']  ; logging.info(msg)

            dummy = self.SetupFileStructure()

            dummy = self.RecoverRun()

            dummy = self.RandomSeed('read')
#
# case 2: A continuation, this is signaled by rc-item time.restart = True
#
        elif self['time.restart']:

            msg   = "Restarting filter from previous step"  ; logging.info(msg)

            dummy = self.SetupFileStructure()

            strippedname = os.path.split(self['jobrcfilename'])[-1]

            self['jobrcfilename'] = os.path.join(self['dir.exec'],strippedname)

            dummy = self.RandomSeed('read')
#
# case 3: A fresh start, this is signaled by rc-item time.restart = False
#
        elif not self['time.restart']:
            msg   = "First time step in filter sequence"   ; logging.info(msg)

            dummy = self.SetupFileStructure()

            # expand jobrcfilename to include exec dir from now on.

            # First strip current leading path from filename

            strippedname = os.path.split(self['jobrcfilename'])[-1]

            self['jobrcfilename'] = os.path.join(self['dir.exec'],strippedname)


        self.ParseTimes()
        self.WriteRC(self['jobrcfilename'])

        return None


    def SetupFileStructure(self):
        """ 
        Create file structure needed for data assimilation system.
        In principle this looks like:

            * ``${da_rundir}``
            * ``${da_rundir}/input``
            * ``${da_rundir}/output``
            * ``${da_rundir}/exec``
            * ``${da_rundir}/diagnostics``
            * ``${da_rundir}/analysis``
            * ``${da_rundir}/jobs``
            * ``${da_rundir}/restart/current``
            * ``${da_rundir}/restart/one-ago``

        .. note:: The exec dir will actually be a simlink to the directory where
                 the observation operator executable lives. This directory is passed through
                 the ``da.rc`` file. 

        .. note:: The observation input files will be placed in the exec dir,
                 and the resulting simulated values will be retrieved from there as well.

        """
        from da.tools.general import CreateDirs

# Create the run directory for this DA job, including I/O structure

        filtertime  = self['time.start'].strftime('%Y%m%d')

        self['dir.exec']        = os.path.join(self['dir.da_run'],'exec')
        self['dir.input']       = os.path.join(self['dir.da_run'],'input')
        self['dir.output']      = os.path.join(self['dir.da_run'],'output',filtertime)
        self['dir.diagnostics'] = os.path.join(self['dir.da_run'],'diagnostics')
        self['dir.analysis']    = os.path.join(self['dir.da_run'],'analysis')
        self['dir.jobs']        = os.path.join(self['dir.da_run'],'jobs')
        self['dir.restart']     = os.path.join(self['dir.da_run'],'restart')
        self['dir.restart.current']    = os.path.join(self['dir.restart'],'current')
        self['dir.restart.oneago']     = os.path.join(self['dir.restart'],'one-ago')

        CreateDirs(self['dir.da_run'])
        CreateDirs(os.path.join(self['dir.exec']))
        CreateDirs(os.path.join(self['dir.input']))
        CreateDirs(os.path.join(self['dir.output']))
        CreateDirs(os.path.join(self['dir.diagnostics']))
        CreateDirs(os.path.join(self['dir.analysis']))
        CreateDirs(os.path.join(self['dir.jobs']))
        CreateDirs(os.path.join(self['dir.restart']))
        CreateDirs(os.path.join(self['dir.restart.current']))
        CreateDirs(os.path.join(self['dir.restart.oneago']))

        msg = 'Succesfully created the file structure for the assimilation job'  ; logging.info(msg)


    def RecoverRun(self):
        """
        Prepare a recovery from a crashed run. This consists of: 
        
            - copying all data from the restart/one-ago folder (:meth:`~da.tools.initexit.CycleControl.MoveRestartData`),
            - replacing all ``rc-file`` items with those from the ``da_runtime.rc`` in the restart/current dir 
            - resetting the seed of the random number generator to the value it had before the crash (:meth:`~da.tools.initexit.CycleControl.RandomSeed`)
            - replacing the output dir name, since it has the sample time in it...
            
        """
        import da.tools.rc as rc
        import shutil
        from da.tools.general import CreateDirs

        # Replace rc-items with those from the crashed run's last rc-file (now in restart.current dir)

        file_rc_rec = os.path.join(self['dir.restart.current'],'da_runtime.rc')
        rc_rec      = rc.read(file_rc_rec)

        for k,v in rc_rec.iteritems():
            self[k] = v

        self.ValidateRC()

        msg         = "Replaced rc-items.... "                            ; logging.debug(msg)
        msg         = "Next cycle start date is %s" % self['time.start']  ; logging.debug(msg)

        # Copy randomseed.pickle file to exec dir

        source        = os.path.join(self['dir.restart.current'],'randomseed.pickle')
        dest          = os.path.join(self['dir.exec'],'randomseed.pickle')
        dummy         = shutil.copy(source,dest)

        msg           = "Replaced randomseed file with previous cycles' last values"          ; logging.debug(msg)

        # Re-create the output dir for this time step, if needed

        filtertime              = self['time.start'].strftime('%Y%m%d')
        self['dir.output']      = os.path.join(self['dir.da_run'],'output',filtertime)
        dummy                   = CreateDirs(os.path.join(self['dir.output']))

        return None

    def Finalize(self):
        """
        Finalize the da cycle, this means writing the save data and rc-files for the next run. 
        The following sequence of actions occur:

            * Write the randomseed to file for reuse in next cycle
            * Write a new ``rc-file`` with ``time.restart : True``, and new ``time.start`` and ``time.end``
            * Collect all needed data needed for check-pointing (restart from current system state)
            * Move the previous check pointing data out of the way, and replace with current
            * Submit the next cycle

        """

        dummy = self.RandomSeed('write')
        dummy = self.WriteNewRCfile()
        dummy = self.MoveRestartData(io_option='store')  # Move restart data from current to one-ago
        dummy = self.CollectRestartData()  # Collect restart data for next cycle into a clean restart/current folder
        dummy = self.CollectOutput()  # Collect restart data for next cycle into a clean restart/current folder
        dummy = self.SubmitNextCycle()

    def CollectOutput(self):
        """ Collect files that are part of the requested output for this cycle. This function allows users to add files 
            to a list, and then the system will copy these to the current cycle's output directory.
            The list of files included is read from the 
            attribute "OutputFileList" which is a simple list of files that can be appended by other objects/methods that
            require output data to be saved.


        """
        from da.tools.general import CreateDirs

        targetdir   = os.path.join(self['dir.output'])

        CreateDirs(os.path.join(targetdir) )

        msg         = "Collecting the required output data"                                       ; logging.info(msg)
        msg         = "           to   directory: %s " % targetdir                                ; logging.debug(msg)

        for file in set(self.OutputFileList):

            if os.path.isdir(file): # skip dirs
                continue
            if not os.path.exists(file): # skip dirs
                msg         = "           [not found] .... %s " % file                                     ; logging.warning(msg)
                continue

            msg         = "           [copy] .... %s " % file                                     ; logging.debug(msg)
            dummy       = shutil.copy(file,file.replace(os.path.split(file)[0],targetdir) )



    def CollectRestartData(self):
        """ Collect files needed for the restart of this cycle in case of a crash, or for the continuation of the next cycle. 
            All files needed are written to the restart/current directory. The list of files included is read from the 
            attribute "RestartFileList" which is a simple list of files that can be appended by other objects/methods that
            require restart data to be saved.

            .. note:: Before collecting the files in the ``RestartFileList``, the restart/current directory will be emptied and
                     recreated. This prevents files from accumulating in the restart/current and restart/one-ago folders. It 
                     also means that if a file is missing from the ``RestartFileList``, it will not be available for check-pointing
                     if your run crashes or dies!

            Currently, the following files are included:

                * The ``da_runtime.rc`` file
                * The ``randomseed.pickle`` file
                * The savestate.nc file
                * The files in the ``ObservationOperator.RestartFileList``, i.e., restart data for the transport model


            .. note:: We assume that the restart files for the :ref:`ObservationOperator` 
                      reside in a separate folder, i.e, the ObservationOperator does *not* write directly to the CTDAS restart dir!

        """
        from da.tools.general import CreateDirs

        targetdir   = os.path.join(self['dir.restart.current'])

        msg         = "Purging the current restart directory before collecting new data"         ; logging.info(msg)

        CreateDirs(os.path.join(targetdir),forceclean=True)

        msg         = "Collecting the required restart data"                                      ; logging.info(msg)
        msg         = "           to   directory: %s " % targetdir                                ; logging.debug(msg)

        for file in set(self.RestartFileList):

            if os.path.isdir(file): # skip dirs
                continue
            if not os.path.exists(file): # skip dirs
                msg         = "           [not found] .... %s " % file                                     ; logging.warning(msg)
            else:
                msg         = "           [copy] .... %s " % file                                     ; logging.debug(msg)
                dummy       = shutil.copy(file,file.replace(os.path.split(file)[0],targetdir) )


    def MoveRestartData(self, io_option='restore'):
        """ 
        Store or restore model state to/from a restart directory. 

            Two IO options are available:

            (1) io_option = restore    : Get data from restart.oneago directory
            (2) io_option = store      : Save data to restart.oneago directory

            In case of a 'store' command the restart.oneago folder is re-created so that the contents are empty to begin with.

        """
        from da.tools.general import CreateDirs

        if io_option not in ['store','restore']:
            raise ValueError,'Invalid option specified for io_option (%s)' % io_option

        if io_option == 'store':

            targetdir = self['dir.restart.oneago']
            sourcedir = self['dir.restart.current']

        elif io_option == 'restore':

            sourcedir = self['dir.restart.oneago']
            targetdir = self['dir.restart.current']

# If "store" is requested, recreate target dir, cleaning the contents 

        if io_option == 'store':
            CreateDirs(os.path.join(targetdir),forceclean=True)

        msg         = "Performing a %s of data" % (io_option)                                     ; logging.debug(msg)
        msg         = "           from directory: %s " % sourcedir                                ; logging.debug(msg)
        msg         = "           to   directory: %s " % targetdir                                ; logging.debug(msg)


        for file in os.listdir(sourcedir):

            file = os.path.join(sourcedir,file)
            if not os.path.exists(file):
                msg         = "Cannot find requested file to move: %s " % file                                ; logging.debug(msg)
                sys.exit(2)


            if os.path.isdir(file): # skip dirs

                msg         = "           [skip] .... %s " % file                                     ; logging.debug(msg)
                continue    
            else:

                msg         = "           [copy] .... %s " % file                                                ; logging.debug(msg)
                dummy       = shutil.copy(file,file.replace(sourcedir,targetdir) )

#
    def WriteNewRCfile(self):
        """ Write the rc-file for the next DA cycle. 

            .. note:: The start time for the next cycle is the end time of this one, while 
                      the end time for the next cycle is the current end time + one cycle length. 
                      
            The resulting rc-file is written to the ``dir.exec`` so that it can be used when resubmitting the next cycle
            
        """
        import da.tools.rc as rc
        import copy

        # We make a copy of the current DaCycle object, and modify the start + end dates and restart value

        newDaCycle                          = copy.deepcopy(self)
        dummy                               = newDaCycle.AdvanceCycleTimes()
        newDaCycle['time.restart']          = True

        # Create the name of the rc-file that will hold this new input, and write it

        fname   = os.path.join(self['dir.exec'],'da_runtime.rc')  # current exec dir holds next rc file
        dummy   = rc.write(fname,newDaCycle)
        msg     = 'Wrote new da_runtime.rc (%s) to exec dir'%fname ; logging.debug(msg)

        # The rest is info needed for a system restart, so it modifies the current DaCycle object (self)

        self['da.restart.fname']             = fname    # needed for next job template
        dummy                                = self.RestartFileList.extend([fname])  # current restart list holds next rc file name

        msg = 'Added da_runtime.rc to the RestartFileList for later collection' ; logging.debug(msg)


    def WriteRC(self,fname):
        """ Write RC file after each process to reflect updated info """
        import da.tools.rc as rc

        dummy  = rc.write(fname,self)
        msg    = 'Wrote expanded rc-file (%s)'%(fname) ; logging.debug(msg)
        return None

    def SubmitNextCycle(self):
        """ 
        Submit the next job of a DA cycle, this consists of 
            * Changing to the working directory from which the job was started initially
            * create a line to start the master script again with a newly created rc-file
            * Submitting the jobfile 

        If the end of the cycle series is reached, no new job is submitted.

        """
        import subprocess
        import os
        from string import join


        DaPlatForm = self.DaPlatForm

        if self['time.end'] < self['time.finish']:

            # file ID and names
            jobid                   = self['time.end'].strftime('%Y%m%d') 
            targetdir               = os.path.join(self['dir.exec'])
            jobfile                 = os.path.join(targetdir,'jb.%s.jb'%jobid)
            logfile                 = jobfile.replace('.jb','.log')

            # Template and commands for job
            jobparams               = {'jobname':"j.%s"%jobid,'jobtime':'01:30:00'}
            template                = DaPlatForm.GetJobTemplate(jobparams)
            execcommand             = os.path.join(self['dir.da_submit'],sys.argv[0]) 
            template                += 'python %s rc=%s %s >& %s' % (execcommand,self['da.restart.fname'],join(self.opts,''),logfile,) 

            # write and submit 

            dummy                   = DaPlatForm.WriteJob(jobfile,template,jobid)
            jobid                   = DaPlatForm.SubmitJob(jobfile,joblog=logfile) 
        else:
            logging.info('Final date reached, no new cycle started')

        return None

    def SubmitSubStep(self,stepname):
        """ 
        Submit the next substep of a DA cycle, this consists of 
            * getting a job template as returned by :meth:`~da.tools.baseclasses.platform.GetJobTemplate`
            * adding the lines needed to start a next run with a newly created rc-file
            * Writing the jobfile as done by :meth:`~da.tools.baseclasses.platform.WriteJob`
            * Submitting the jobfile as done by :meth:`~da.tools.baseclasses.platform.WriteJob`

        """
        import subprocess
        import os
        from string import join


        DaPlatForm = self.DaPlatForm

        jobparams               = {'jobname':'das.%s'%stepname}
        template                = DaPlatForm.GetJobTemplate(jobparams)
        template                += 'cd %s\n'%os.getcwd()
        template                += '%s rc=%s process=%s %s' % (sys.argv[0],self['jobrcfilename'],stepname,join(self.opts,''),) 
        jobfile                 = DaPlatForm.WriteJob(self,template,stepname)
        jobid                   = DaPlatForm.SubmitJob(jobfile) 

        return None

    def CleanUpCycle(self):
        """
        Nothing to do for now anymore
        """



def StartLogger(level=logging.INFO):
    """ start the logging of messages to screen"""

# start the logging basic configuration by setting up a log file

    logging.basicConfig(level    = level,
                        format   = ' [%(levelname)-7s] (%(asctime)s) py-%(module)-20s : %(message)s',
                        datefmt  = '%Y-%m-%d %H:%M:%S')

def ParseOptions():
    """ 
    Function parses options from the command line and returns the arguments as a dictionary.
    Accepted command line arguments are:

    ========  =======
    Argument  Meaning
    ========  =======
    -v        verbose output in log files
    -h        display help
    -r        start a simulation by recovering from a previous crash
    ========  =======

    """
    import getopt
    import sys

# Parse keywords, the only option accepted so far is the "-h" flag for help

    opts=[]
    args=[]
    try:                                
        opts, args = getopt.gnu_getopt(sys.argv[1:], "-hrv")
    except getopt.GetoptError, msg:           
        logging.error('%s'%msg)
        sys.exit(2)      

    for options in opts:
        options=options[0].lower()
        if options == '-h':
            print ""
            print helptext
            sys.exit(2)      
        if options == '-r':
            logging.info('-r flag specified on command line: recovering from crash')
        if options == '-v':
            logging.info('-v flag specified on command line: extra verbose output')
            dummy  = logging.root.setLevel(logging.DEBUG)

    if opts: 
        optslist=[item[0] for item in opts]
    else:
        optslist=[]

# Parse arguments and return as dictionary

    arguments={}
    for item in args:
        #item=item.lower()

# Catch arguments that are passed not in "key=value" format

        if '=' in item:
            key, arg = item.split('=')
        else:
            logging.error('%s'%'Argument passed without description (%s)' % item)
            raise getopt.GetoptError,arg

        arguments[key]=arg


    return optslist, arguments

def ValidateOptsArgs(opts,args):
    """ 
 Validate the options and arguments passed from the command line before starting the cycle. The validation consists of checking for the presence of an argument "rc", and the existence of
 the specified rc-file.  
 
    """

    if not args.has_key("rc"):
        msg = "There is no rc-file specified on the command line. Please use rc=yourfile.rc"   ; logging.error(msg)
        raise IOError,msg
    elif not os.path.exists(args['rc']):
        msg = "The specified rc-file (%s) does not exist " % args['rc'] ;  logging.error(msg)
        raise IOError,msg

    # WP not needed anymore
    #if not args.has_key('process'):
    #    msg = "There is no process specified on the command line, assuming process=Start"   ; logging.info(msg)
    #    args['process'] = 'start'
    #if args['process'].lower() not in validprocesses:
    #    msg = "The specified process (%s) is not valid"%args['process']   ; logging.error(msg)
    #    raise IOError,msg

    return opts,args


if __name__ == "__main__":

    sys.path.append('../../')
    opts,args = ParseOptions()
    print opts
    print args

