#!/usr/bin/env python
# pipeline.py

"""
.. module:: pipeline
.. moduleauthor:: Wouter Peters 

Revision History:
File created on 06 Sep 2010.

The pipeline module holds methods that execute consecutive tasks with each of the objects of the DA system. 

"""
import logging
import os
import sys
import shutil
import datetime
import da.tools.rc as rc

header  = '\n\n    ***************************************   '
footer  = '    *************************************** \n  '

def EnsembleSmootherPipeline(DaCycle,PlatForm, DaSystem, Samples,StateVector,ObsOperator,Optimizer):
    """ The main point of entry for the pipeline """

# Append current working dir to path

    sys.path.append(os.getcwd())

# Import methods and classes contained in this package

    from da.tools.general import StoreData,RestoreData
    
    msg          = header+"Initializing current cycle"+footer           ; logging.info(msg) 

    status        = JobStart(DaCycle, DaSystem, PlatForm, StateVector, Samples, ObsOperator)

    msg          = header+"starting PrepareState"+footer           ; logging.info(msg) 

    status        = PrepareState(DaCycle, StateVector)

    msg          = header+"starting SampleState"+footer             ; logging.info(msg) 

    status        = SampleState(DaCycle,Samples,StateVector,ObsOperator)

    msg          = header+"starting Invert"+footer             ; logging.info(msg) 

    status        = Invert(DaCycle, StateVector, Optimizer)

    msg          = header+"starting Advance"+footer          ; logging.info(msg) 

    status        = Advance(DaCycle, Samples, StateVector, ObsOperator)

    msg          = header+"starting SaveAndSubmit"+footer      ; logging.info(msg) 

    status        = SaveAndSubmit(DaCycle, StateVector)

    msg          = "Cycle finished...exiting pipeline"     ; logging.info(msg)

####################################################################################################

def JobStart(DaCycle, DaSystem, DaPlatForm, StateVector, Samples, ObsOperator):
    """ Set up the job specific directory structure and create an expanded rc-file """

    status               = DaSystem.Initialize()

    status               = DaSystem.Validate()

    DaCycle.DaSystem    = DaSystem

    DaCycle.DaPlatForm  = DaPlatForm

    status               = DaCycle.Initialize()

    StateVector.DaCycle = DaCycle # also embed object in StateVector so it can access cycle information for I/O etc

    Samples.DaCycle     = DaCycle # also embed object in Samples object so it can access cycle information for I/O etc

    ObsOperator.DaCycle = DaCycle # also embed object in ObsOperator object so it can access cycle information for I/O etc

    status              = ObsOperator.Initialize()  # Setup Observation Operator 

    return None


def PrepareState(DaCycle, StateVector):
    """ Set up the input data for the forward model: obs and parameters/fluxes"""

    # We now have an empty StateVector object that we need to populate with data. If this is a continuation from a previous cycle, we can read
    # the previous StateVector values from a NetCDF file in the restart/current directory. If this is the first cycle, we need to populate the StateVector
    # with new values for each week. After we have constructed the StateVector, it will be propagated by one cycle length so it is ready to be used
    # in the current cycle

    status               = StateVector.Initialize()

    if not DaCycle['time.restart']:

        # Fill each week from n=1 to n=nlag with a new ensemble

        nlag = StateVector.nlag
        for n in range(0,nlag):
            date = DaCycle['time.start']+datetime.timedelta(days = (n+0.5)* int(DaCycle['time.cycle']))
            cov   = StateVector.GetCovariance(date)
            status = StateVector.MakeNewEnsemble(n+1,cov)

    else:

        # Read the StateVector data from file

        savedir         = DaCycle['dir.restart.current']
        filtertime      = DaCycle['time.start'].strftime('%Y%m%d')
        filename        = os.path.join(savedir,'savestate.nc')

        StateVector.ReadFromFile(filename) # by default will read "opt"(imized) variables, and then propagate

        # Now propagate the ensemble by one cycle to prepare for the current cycle

        status = StateVector.Propagate()

   # Finally, also write the StateVector to a file so that we can always access the a-priori information

    savedir         = DaCycle['dir.output']
    filename        = os.path.join(savedir,'savestate.nc')

    status = StateVector.WriteToFile(filename)  # write prior info because StateVector.Isoptimized == False for now

    return None

def SampleState(DaCycle,Samples,StateVector, ObservationOperator):
    """ Sample the filter state for the inversion """


    # Before a forecast step, save all the data to a save/tmp directory so we can later recover it before the propagation step.
    # This is especially important for:
    #  (i) The transport model restart data which holds the background mixing ratios. This is needed to run the model one cycle forward
    #  (ii) The random numbers (or the seed for the random number generator) so we can recreate the ensembles if needed

    #status   = DaCycle.MoveSaveData(io_option='store',save_option='partial',filter=[])
    #msg     = "All restart data have been copied to the save/tmp directory for future use"    ; logging.debug(msg)

    nlag    = int(DaCycle['time.nlag'])
    msg     = "Sampling model will be run over %d cycles"% nlag ; logging.info(msg)

    for lag in range(nlag):

        msg     = header+".....Ensemble Kalman Filter at lag %d"% (int(lag)+1,)           ; logging.info(msg) 

    # If we are going to sample the lag = 0 time step, place the needed files to run the transport model in the right folder first

        if lag == 0:
            status  = ObservationOperator.GetInitialData()

        ############# Perform the actual sampling loop #####################

        status      = SampleOneCycle(DaCycle,Samples,StateVector, ObservationOperator,lag)

        ############# Finished the actual sampling loop #####################

        msg = "StateVector now carries %d samples"%StateVector.nobs ; logging.debug(msg)

        # Add the observation operator restart+output files to the general Output+RestartFileList, only after 'Advance'
        # Same logic for observation files

        if lag == 0:
            DaCycle.RestartFileList.extend( ObservationOperator.RestartFileList )
            DaCycle.OutputFileList.extend(ObservationOperator.OutputFileList )
            msg = "Appended ObsOperator restart and output file lists to DaCycle for collection " ; logging.debug(msg)
        
            DaCycle.OutputFileList.extend([DaCycle['ObsOperator.inputfile']])
            msg = "Appended Observation filename to DaCycle for collection " ; logging.debug(msg)


    # Optionally, post-processing of the model output can be added that deals for instance with
    # sub-sampling of time series, vertical averaging, etc.

    return None

def SampleOneCycle(DaCycle,Samples,StateVector, ObservationOperator,lag):
    """ Perform all actions needed to sample one cycle """
    import copy

    # First set up the information for time start and time end of this sample


    status                          = DaCycle.SetSampleTimes(lag)

    startdate                       = DaCycle['time.sample.start'] 
    enddate                         = DaCycle['time.sample.end'] 
    DaCycle['time.sample.window']   = lag
    DaCycle['time.sample.stamp']    = "%s_%s"%(startdate.strftime("%Y%m%d%H"),enddate.strftime("%Y%m%d%H"),)

    msg   = "New simulation interval set : "                                       ; logging.info(msg)
    msg   = "                  start date : %s " % startdate.strftime('%F %H:%M')  ; logging.info(msg)
    msg   = "                  end   date : %s " % enddate.strftime('%F %H:%M')    ; logging.info(msg)
    msg   = "                  file  stamp: %s " % DaCycle['time.sample.stamp']    ; logging.info(msg)


    # Implement something that writes the ensemble member parameter info to file, or manipulates them further into the 
    # type of info needed in your transport model

    status   = StateVector.WriteMembersToFile(lag+1)

    status   = Samples.Initialize()

    status   = Samples.Validate() 

    status   = Samples.AddObs() 

    filename   = Samples.WriteSampleInfo() 

    # Write filename to DaCycle, and to output collection list

    DaCycle['ObsOperator.inputfile'] = filename

    # Run the observation operator

    status   = RunForecastModel(DaCycle,ObservationOperator)

    # Add model-data mismatch to all samples, this *might* use output from the ensemble in the future??

    status   = Samples.AddModelDataMismatch()

    msg = "Added Model Data Mismatch to all samples " ; logging.debug(msg)

    # Read forecast model samples that were written to NetCDF files by each member. Add them to the exisiting
    # Observation object for each sample loop. This data fill be written to file in the output folder for each sample cycle. 
    

    # We retrieve all model samples from one output file written by the ObsOperator. If the ObsOperator creates
    # one file per member, some logic needs to be included to merge all files!!!
         
    filename        = os.path.join(ObservationOperator.outputdir,'flask_output.%s.nc'%DaCycle['time.sample.stamp'])
    status          = Samples.AddSimulations(filename)

    # Now write a small file that holds for each observation a number of values that we need in postprocessing

    filename        = Samples.WriteSampleInfo() 

    # Now add the observations that need to be assimilated to the StateVector. 
    # Note that obs will only be added to the statevector if either this is the first step (restart=False), or lag==nlag
    # This is to make sure that the first step of the system uses all observations available, while the subsequent
    # steps only optimize against the data at the front (lag==nlag) of the filter. This way, each observation is used only 
    # (and at least) once # in the assimilation

    if lag == int(DaCycle['time.nlag'])-1 or DaCycle['time.restart']==False :

        StateVector.ObsToAssimmilate += (copy.deepcopy(Samples),)

        StateVector.nobs += Samples.getlength()

        msg = "Added samples from the observation operator to the assimilated obs list in the StateVector" ; logging.debug(msg)

    else:

        StateVector.ObsToAssimmilate += (None,)


    return None

def Invert(DaCycle, StateVector, Optimizer ):
    """ Perform the inverse calculation """

    dims      = ( int(DaCycle['time.nlag']),
                  int(DaCycle['da.optimizer.nmembers']),
                  int(DaCycle.DaSystem['nparameters']),
                  StateVector.nobs,  )

    status     = Optimizer.Initialize(dims)

    status     = Optimizer.StateToMatrix(StateVector )

    status     = Optimizer.WriteDiagnostics(DaCycle,StateVector,type='prior')

    status     = Optimizer.SetLocalization('None')

    if not DaCycle.DaSystem.has_key('opt.algorithm'):
        msg       = "There was no minimum least squares algorithm specified in the DA System rc file (key : opt.algorithm)" ; logging.info(msg) 
        msg       = "...using serial algorithm as default..." ; logging.info(msg)
        status     = Optimizer.SerialMinimumLeastSquares()
    elif DaCycle.DaSystem['opt.algorithm'] == 'serial':
        msg       = "Using the serial minimum least squares algorithm to solve ENKF equations" ; logging.info(msg)
        status     = Optimizer.SerialMinimumLeastSquares()
    elif DaCycle.DaSystem['opt.algorithm'] == 'bulk':
        msg       = "Using the bulk minimum least squares algorithm to solve ENKF equations" ; logging.info(msg)
        status     = Optimizer.BulkMinimumLeastSquares()


    status     = Optimizer.MatrixToState(StateVector )

    status     = Optimizer.WriteDiagnostics(DaCycle,StateVector,type='optimized')

    StateVector.isOptimized = True

    return None

def Advance( DaCycle, Samples, StateVector, ObservationOperator):
    """ Advance the filter state to the next step """

    # This is the advance of the modeled CO2 state. Optionally, routines can be added to advance the state vector (mean+covariance)

    # Then, restore model state from the start of the filter

    msg     = "Sampling model will be run over 1 cycle"  ; logging.info(msg)

    SampleOneCycle(DaCycle,Samples,StateVector, ObservationOperator,0)

    # Write info from optimized flux cycle to a file

    filename = Samples.WriteObsToFile()

    return None

def SaveAndSubmit( DaCycle, StateVector):
    """ Save the model state and submit the next job """

    savedir         = DaCycle['dir.output']
    filename        = os.path.join(savedir,'savestate.nc')

    status = StateVector.WriteToFile(filename)

    DaCycle.RestartFileList.extend( [filename] ) # write optimized info because StateVector.Isoptimized == False for now


    status = DaCycle.Finalize()

    return None

def RunForecastModel(DaCycle,ObsOperator):
    """Prepare and execute a forecast step using an external Fortran model. Note that the flavor of model 
       used is determined in the very first line where the import statement of module "model" depends on a 
       setting in your da.rc file. After that choice, the module written specifically for a particular 
       application takes over. There are a few required variables and methods in such a module:
       
       version [variable, string]    : defines the version number of the module, e.g. "1.0"
       identifier [variable, string] : identifies the model used as a string variable, e.g. "TM5"
       PrepareExe (method)           : A method that preps the model simulation. It can for instance create an input 
                                       parameter file for the model (tm5.rc), or execute some scripts needed before the 
                                       actual execution starts (get_meteo). After this step, a call to the model 
                                       executable should start a succesfull simulation
       Run        (method)           : Start the executable. How this is done depends on your model and might involve 
                                       submitting a string of jobs to the queue, or simply spawning a subprocess, or ...
       """

    status = ObsOperator.PrepareRun()

    status = ObsOperator.ValidateInput()

    status = ObsOperator.Run()

    status = ObsOperator.SaveData()

    return status



